"""HMC7044 clock chip model."""

from typing import Dict, List, Union

from adijif.clocks.hmc7044_bf import hmc7044_bf

from adijif.solvers import CpoExpr, CpoModel  # type: ignore # isort: skip  # noqa: I202
from adijif.solvers import CpoSolveResult  # type: ignore # isort: skip  # noqa: I202
from adijif.solvers import GEKKO  # type: ignore # isort: skip  # noqa: I202
from adijif.solvers import GK_Intermediate  # type: ignore # isort: skip  # noqa: I202

from adijif.draw import Layout, Node  # type: ignore # isort: skip  # noqa: I202


class hmc7044(hmc7044_bf):
    """HMC7044 clock chip model.

    This model currently supports VCXO+PLL2 configurations
    """

    name = "HMC7044"

    # Ranges
    # r2_divider_min = 1
    # r2_divider_max = 4095
    r2_available = [*range(1, 4095 + 1)]

    """ Output dividers """
    d_available = [1, 3, 5, *range(2, 4095, 2)]
    # When pulse generation is required (like for sysref) divder range
    # is limited
    d_syspulse_available = [*range(32, 4095, 2)]

    # Defaults
    _d: Union[int, List[int]] = [1, 3, 5, *range(2, 4095, 2)]
    _r2: Union[int, List[int]] = [*range(1, 4095 + 1)]

    # Limits
    """ Internal limits """
    vco_min = 2400e6
    vco_max = 3200e6
    pfd_max = 250e6
    vcxo_min = 10e6
    vcxo_max = 500e6

    use_vcxo_double = True
    vxco_doubler_available = [1, 2]
    _vcxo_doubler = [1, 2]

    minimize_feedback_dividers = True

    # State management
    _clk_names: List[str] = []

    def __init__(
        self, model: Union[GEKKO, CpoModel] = None, solver: str = "CPLEX"
    ) -> None:
        """Initialize HMC7044 clock chip model.

        Args:
            model (Model): Model to add constraints to
            solver (str): Solver to use. Should be one of "CPLEX" or "gekko"

        Raises:
            Exception: Invalid solver
        """
        super(hmc7044, self).__init__(model, solver)
        if solver == "gekko":
            self.n2_available = [*range(8, 65535 + 1)]
            self._n2 = [*range(8, 65535 + 1)]
        elif solver == "CPLEX":
            self.n2_available = [*range(8, 65535 + 1)]
            self._n2 = [*range(8, 65535 + 1)]
        else:
            raise Exception("Unknown solver {}".format(solver))

    @property
    def d(self) -> Union[int, List[int]]:
        """Output dividers.

        Valid dividers are 1,2,3,4,5,6->(even)->4094

        Returns:
            int: Current allowable dividers
        """
        return self._d

    @d.setter
    def d(self, value: Union[int, List[int]]) -> None:
        """Output dividers.

        Valid dividers are 1,2,3,4,5,6->(even)->4094

        Args:
            value (int, list[int]): Allowable values for divider
        """
        self._check_in_range(value, self.d_available, "d")
        self._d = value

    @property
    def n2(self) -> Union[int, List[int]]:
        """n2: VCO feedback divider.

        Valid dividers are 8->65536

        Returns:
            int: Current allowable dividers
        """
        return self._n2

    @n2.setter
    def n2(self, value: Union[int, List[int]]) -> None:
        """VCO feedback divider.

        Valid dividers are 8->65536

        Args:
            value (int, list[int]): Allowable values for divider
        """
        self._check_in_range(value, self.n2_available, "n2")
        self._n2 = value

    @property
    def r2(self) -> Union[int, List[int]]:
        """VCXO input dividers.

        Valid dividers are 1->4096

        Returns:
            int: Current allowable dividers
        """
        return self._r2

    @r2.setter
    def r2(self, value: Union[int, List[int]]) -> None:
        """VCXO input dividers.

        Valid dividers are 1->4096

        Args:
            value (int, list[int]): Allowable values for divider
        """
        self._check_in_range(value, self.r2_available, "r2")
        self._r2 = value

    @property
    def vxco_doubler(self) -> Union[int, List[int]]:
        """VCXO doubler.

        Valid dividers are 1,2

        Returns:
            int: Current doubler value
        """
        return self._vcxo_doubler

    @vxco_doubler.setter
    def vxco_doubler(self, value: Union[int, List[int]]) -> None:
        """VCXO doubler.

        Valid dividers are 1,2

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.vxco_doubler_available, "vxco_doubler")
        self._vcxo_doubler = value

    def _init_diagram(self) -> None:
        """Initialize diagram for HMC7044 alone."""
        self.ic_diagram_node = None
        self._diagram_output_dividers = []

        # lo = Layout("HMC7044 Example")

        self.ic_diagram_node = Node("HMC7044")
        # lo.add_node(root)

        # External
        # ref_in = Node("REF_IN", ntype="input")
        # lo.add_node(ref_in)

        vcxo_doubler = Node("VCXO Doubler", ntype="shell")
        self.ic_diagram_node.add_child(vcxo_doubler)

        # Inside the IC
        r2_div = Node("R2", ntype="divider")
        # r2_div.value = "2"
        self.ic_diagram_node.add_child(r2_div)
        pfd = Node("PFD", ntype="phase-frequency-detector")
        self.ic_diagram_node.add_child(pfd)
        lf = Node("LF", ntype="loop-filter")
        self.ic_diagram_node.add_child(lf)
        vco = Node("VCO", ntype="voltage-controlled-oscillator")
        vco.shape = "circle"
        self.ic_diagram_node.add_child(vco)
        n2 = Node("N2", ntype="divider")
        self.ic_diagram_node.add_child(n2)

        out_dividers = Node("Output Dividers", ntype="shell")
        # ds = 4
        # out_divs = []
        # for i in range(ds):
        #     div = Node(f"D{i+1}", ntype="divider")
        #     out_dividers.add_child(div)
        #     out_divs.append(div)

        self.ic_diagram_node.add_child(out_dividers)

        # Connections inside the IC
        # lo.add_connection({"from": ref_in, "to": r2_div, 'rate': 125000000})
        self.ic_diagram_node.add_connection({"from": vcxo_doubler, "to": r2_div})
        self.ic_diagram_node.add_connection(
            {"from": r2_div, "to": pfd, "rate": 125000000 / 2}
        )
        self.ic_diagram_node.add_connection({"from": pfd, "to": lf})
        self.ic_diagram_node.add_connection({"from": lf, "to": vco})
        self.ic_diagram_node.add_connection({"from": vco, "to": n2})
        self.ic_diagram_node.add_connection({"from": n2, "to": pfd})

        self.ic_diagram_node.add_connection(
            {"from": vco, "to": out_dividers, "rate": 4000000000}
        )
        # for div in out_divs:
        #     self.ic_diagram_node.add_connection({"from": out_dividers, "to": div})
        #     # root.add_connection({"from": vco, "to": div})

    def _update_diagram(self, config: Dict) -> None:
        """Update diagram with configuration.

        Args:
            config (Dict): Configuration dictionary

        Raises:
            Exception: If key is not D followed by a number
        """
        # Add output dividers
        keys = config.keys()
        output_dividers = self.ic_diagram_node.get_child("Output Dividers")
        for key in keys:
            if key.startswith("D"):
                div = Node(key, ntype="divider")
                output_dividers.add_child(div)
                self.ic_diagram_node.add_connection(
                    {"from": output_dividers, "to": div}
                )
            else:
                raise Exception(
                    f"Unknown key {key}. Must be of for DX where X is a number"
                )

    def draw(self, lo: Layout = None) -> str:
        """Draw diagram in d2 language for IC alone with reference clock.

        Args:
            lo: Layout for drawing

        Returns:
            str: Diagram in d2 language

        Raises:
            Exception: If no solution is saved
        """
        if not self._saved_solution:
            raise Exception("No solution to draw. Must call solve first.")

        system_draw = lo is not None
        if not system_draw:
            lo = Layout("HMC7044 Example")
        else:
            # Verify lo is a Layout object
            assert isinstance(lo, Layout), "lo must be a Layout object"
        lo.add_node(self.ic_diagram_node)

        ref_in = Node("REF_IN", ntype="input")
        lo.add_node(ref_in)
        vcxo_double = self.ic_diagram_node.get_child("VCXO Doubler")
        lo.add_connection(
            {"from": ref_in, "to": vcxo_double, "rate": self._saved_solution["vcxo"]}
        )

        # Update Node values
        node = self.ic_diagram_node.get_child("VCXO Doubler")
        node.value = str(self._saved_solution["vcxo_doubler"])
        node = self.ic_diagram_node.get_child("R2")
        node.value = str(self._saved_solution["r2"])
        node = self.ic_diagram_node.get_child("N2")
        node.value = str(self._saved_solution["n2"])

        # Update VCXO Doubler to R2
        # con = self.ic_diagram_node.get_connection("VCXO Doubler", "R2")
        rate = self._saved_solution["vcxo_doubler"] * self._saved_solution["vcxo"]
        self.ic_diagram_node.update_connection("VCXO Doubler", "R2", rate)

        # Update R2 to PFD
        # con = self.ic_diagram_node.get_connection("R2", "PFD")
        rate = (
            self._saved_solution["vcxo"]
            * self._saved_solution["vcxo_doubler"]
            / self._saved_solution["r2"]
        )
        self.ic_diagram_node.update_connection("R2", "PFD", rate)

        # Update VCO
        # con = self.ic_diagram_node.get_connection("VCO", "Output Dividers")
        self.ic_diagram_node.update_connection(
            "VCO", "Output Dividers", self._saved_solution["vco"]
        )

        # Update diagram with dividers and rates
        d = 0
        output_dividers = self.ic_diagram_node.get_child("Output Dividers")

        for key, val in self._saved_solution["output_clocks"].items():
            clk_node = Node(key, ntype="divider")
            div_value = val["divider"]
            div = output_dividers.get_child(f"D{d}")
            div.value = str(div_value)
            d += 1
            lo.add_node(clk_node)
            lo.add_connection({"from": div, "to": clk_node, "rate": val["rate"]})

        if system_draw:
            return lo.draw()

        return lo.draw()

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")

        if solution:
            self.solution = solution

        out_dividers = [self._get_val(x) for x in self.config["out_dividers"]]

        config: Dict = {
            "r2": self._get_val(self.config["r2"]),
            "n2": self._get_val(self.config["n2"]),
            "out_dividers": out_dividers,
            "output_clocks": [],
        }

        # Handle different vcxo types
        if hasattr(self, "vcxo_arb") and self.vcxo_arb:
            # arb_source case
            vcxo_cfg = self.vcxo_arb.get_config(self.solution)  # type: ignore
            vcxo = list(vcxo_cfg.values())[0]
            self.vcxo = vcxo
        elif self.vcxo_i:
            # range case
            vcxo = self._get_val(self.vcxo_i["range"])
            self.vcxo = vcxo

        clk = self.vcxo / config["r2"] * config["n2"]

        output_cfg = {}
        vd = self._get_val(self.config["vcxo_doubler"])
        for i, div in enumerate(out_dividers):
            rate = vd * clk / div
            output_cfg[self._clk_names[i]] = {"rate": rate, "divider": div}

        config["output_clocks"] = output_cfg
        config["vco"] = clk * vd
        config["vcxo"] = self.vcxo
        config["vcxo_doubler"] = vd

        self._saved_solution = config

        return config

    def _setup_solver_constraints(self, vcxo: int) -> None:
        """Apply constraints to solver model.

        Args:
            vcxo (int): VCXO frequency in hertz

        Raises:
            Exception: Invalid solver
        """
        self.vcxo = vcxo

        if self.solver == "gekko":
            self.config = {"r2": self.model.Var(integer=True, lb=1, ub=4095, value=1)}
            self.config["n2"] = self.model.Var(integer=True, lb=8, ub=4095)
            if isinstance(vcxo, (int, float)):
                vcxo_var = self.model.Const(int(vcxo))
            else:
                vcxo_var = vcxo
            self.config["vcxo_doubler"] = self.model.sos1([1, 2])
            self.config["vcxod"] = self.model.Intermediate(
                self.config["vcxo_doubler"] * vcxo_var
            )
        elif self.solver == "CPLEX":
            self.config = {
                "r2": self._convert_input(self._r2, "r2"),
                "n2": self._convert_input(self._n2, "n2"),
            }
            self.config["vcxo_doubler"] = self._convert_input(
                self._vcxo_doubler, "vcxo_doubler"
            )
            self.config["vcxod"] = self._add_intermediate(
                self.config["vcxo_doubler"] * vcxo
            )
        else:
            raise Exception("Unknown solver {}".format(self.solver))

        # PLL2 equations
        self._add_equation(
            [
                self.config["vcxod"] <= self.pfd_max * self.config["r2"],
                self.config["vcxod"] * self.config["n2"]
                <= self.vco_max * self.config["r2"],
                self.config["vcxod"] * self.config["n2"]
                >= self.vco_min * self.config["r2"],
            ]
        )

        # Objectives
        if self.minimize_feedback_dividers:
            if self.solver == "CPLEX":
                self._add_objective(self.config["r2"])
                # self.model.minimize(self.config["r2"])
            elif self.solver == "gekko":
                self.model.Obj(self.config["r2"])
            else:
                raise Exception("Unknown solver {}".format(self.solver))

    def _setup(self, vcxo: int) -> None:
        # Setup clock chip internal constraints

        # FIXME: ADD SPLIT m1 configuration support

        # Convert VCXO into intermediate in case we have range or arb_source type
        if type(vcxo) not in [int, float]:
            vcxo_result = vcxo(self.model)
            # Handle range type (returns dict with "range" key)
            if isinstance(vcxo_result, dict):
                self.vcxo_i = vcxo_result
                self.vcxo_arb = None
                vcxo = self.vcxo_i["range"]
            # Handle arb_source type (returns direct expression)
            else:
                self.vcxo_i = False
                self.vcxo_arb = vcxo  # Store original for get_config
                vcxo = vcxo_result
        else:
            self.vcxo_i = False
            self.vcxo_arb = None

        self._setup_solver_constraints(vcxo)

        # Add requested clocks to output constraints
        self.config["out_dividers"] = []
        self._clk_names = []  # Reset

    def _get_clock_constraint(
        self, clk_name: List[str]
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  String of clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference

        Raises:
            Exception: Invalid solver
        """
        if self.solver == "gekko":

            __d = self._d if isinstance(self._d, list) else [self._d]

            if __d.sort() != self.d_available.sort():
                raise Exception("For solver gekko d is not configurable for HMC7044")

            even = self.model.Var(integer=True, lb=3, ub=2047)
            odd = self.model.Intermediate(even * 2)
            od = self.model.sos1([1, 2, 3, 4, 5, odd])

        elif self.solver == "CPLEX":
            od = self._convert_input(self._d, "d_" + str(clk_name))
        else:
            raise Exception("Unknown solver {}".format(self.solver))

        # Update diagram to include new divider
        d_n = len(self.config["out_dividers"])
        self._update_diagram({f"D{d_n}": od})

        self._clk_names.append(clk_name)

        self.config["out_dividers"].append(od)
        return self.config["vcxod"] / self.config["r2"] * self.config["n2"] / od

    def set_requested_clocks(
        self, vcxo: int, out_freqs: List, clk_names: List[str]
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            vcxo (int): VCXO frequency in hertz
            out_freqs (List): list of required clocks to be output
            clk_names (List[str]):  list of strings of clock names

        Raises:
            Exception: If len(out_freqs) != len(clk_names)
        """
        if len(clk_names) != len(out_freqs):
            raise Exception("clk_names is not the same size as out_freqs")

        # Setup clock chip internal constraints
        self._setup(vcxo)
        self._clk_names = clk_names
        # if type(self.vcxo) not in [int,float]:
        #     vcxo = self.vcxo['range']

        self._saved_solution = None

        # Add requested clocks to output constraints
        for d_n, out_freq in enumerate(out_freqs):

            if self.solver == "gekko":
                __d = self._d if isinstance(self._d, list) else [self._d]
                if __d.sort() != self.d_available.sort():
                    raise Exception(
                        "For solver gekko d is not configurable for HMC7044"
                    )

                # even = self.model.Var(integer=True, lb=3, ub=2047)
                # odd = self.model.Intermediate(even * 2)
                # od = self.model.sos1([1, 2, 3, 4, 5, odd])

                # Since d is so disjoint it is very annoying to solve.
                even = self.model.Var(integer=True, lb=1, ub=4094 // 2)

                # odd = self.model.sos1([1, 3, 5])
                odd_i = self.model.Var(integer=True, lb=0, ub=2)
                odd = self.model.Intermediate(1 + odd_i * 2)

                eo = self.model.Var(integer=True, lb=0, ub=1)
                od = self.model.Intermediate(eo * odd + (1 - eo) * even * 2)

            elif self.solver == "CPLEX":
                od = self._convert_input(self._d, f"d_{out_freq}_{d_n}")

            self._add_equation(
                [
                    self.config["vcxod"] * self.config["n2"]
                    == out_freq * self.config["r2"] * od
                ]
            )
            self.config["out_dividers"].append(od)

            # Update diagram to include new divider
            self._update_diagram({f"D{d_n}": od})

            # Objectives
            # self.model.Obj(-1*eo) # Favor even dividers



"""AD9084 MxFE Utility Functions."""

import json
import os
from typing import Dict, Union

import pandas as pd

from ..utils import get_jesd_mode_from_params
from .converter import converter


def _convert_to_config(
    mode: str,
    L: Union[int, float],
    M: Union[int, float],
    F: Union[int, float],
    S: Union[int, float],
    HD: Union[int, float],
    K: Union[int, float],
    N: Union[int, float],
    Np: Union[int, float],
    CS: Union[int, float],
    E: Union[int, float],
    global_index: Union[int, float],
    jesd_class: str,
) -> Dict:
    return {
        "L": L,
        "M": M,
        "F": F,
        "S": S,
        # "HD": 1 if F == 1 else 0,
        "Np": Np,
        "K": K,
        "HD": HD,
        "CS": CS,
        "E": E,
        "jesd_class": jesd_class,
        "global_index": global_index,
    }


def _load_rx_config_modes(part: str) -> Dict:
    """Load RX JESD configuration tables from file.

    Args:
        part (str): Part name, either "AD9084" or "AD9088".

    Returns:
        Dict: Dictionary of JESD configuration modes.

    Raises:
        AssertionError: If the part is not supported.
    """
    assert part in ["AD9084", "AD9088"], f"Unsupported part: {part}"
    return _read_table_xlsx("AD9084_JTX_JRX.xlsx", part)


def _read_table_xlsx(filename: str, part: str) -> Dict:
    loc = os.path.dirname(__file__)
    fn = os.path.join(loc, "resources", filename)
    table = pd.read_excel(open(fn, "rb"), sheet_name="JTX_RxPath")

    # strip out unique JESD modes
    # table = table.drop_duplicates(subset=["JTX_MODE NUMBER"])
    jrx_modes_204b = {}
    jrx_modes_204c = {}
    for prow in table.iterrows():
        row = prow[1].to_dict()

        field = "8T8R" if part == "AD9088" else "4T4R"
        data = str(row[field])

        if "nan" in data:
            continue
        if "Not Supported" in data:
            continue

        jrx_modes_204c[str(row["Mode"])] = {
            "L": row["L"],
            "M": row["M"],
            "F": row["F"],
            "S": row["S"],
            "HD": 1,
            # 'K': row['K'],
            "Np": row["Np"],
            "DL": row["DL"],
            "jesd_class": "jesd204c",
        }

    # Copy settings to 204b
    for mode, config in jrx_modes_204c.items():
        jrx_modes_204b[mode] = config.copy()
        jrx_modes_204b[mode]["jesd_class"] = "jesd204b"
        # jrx_modes_204b[mode]["HD"] = 0  # HD is always 0 for 204b

    return {"jesd204b": jrx_modes_204b, "jesd204c": jrx_modes_204c}


def parse_json_config(profile_json: str, bypass_version_check: bool = False) -> Dict:
    """Parse Apollo profiles and extract desired part information.

    Args:
        profile_json (str): Path to the profile JSON file.
        bypass_version_check (bool): If True, bypasses the version check for
            the profile.

    Returns:
        Dict: A dictionary containing parsed configuration data.

    Raises:
        FileNotFoundError: If the summary or profile JSON file does not exist.
        KeyError: If required keys are missing in the JSON data.
        Exception: If the profile is not supported or if it is a JESD204B profile.
    """
    use_summary = False  # cannot use summary for now as its broken in ACE
    summary_json = None  # Needed for lint
    if use_summary:
        if not os.path.exists(summary_json):
            raise FileNotFoundError(f"Summary JSON file does not exist: {summary_json}")
    if not os.path.exists(profile_json):
        raise FileNotFoundError(f"Profile JSON file does not exist: {profile_json}")
    full_profile_filename = os.path.abspath(profile_json)

    with open(full_profile_filename, "r") as f:
        profile_data = json.load(f)

    # Check version
    if not bypass_version_check:
        if (
            "profile_cfg" not in profile_data
            or "profile_version" not in profile_data["profile_cfg"]
        ):
            raise KeyError(
                f"ERROR {profile_json} because 'profile_cfg' "
                + f"key is missing in {profile_json}"
            )

        profile_version = profile_data["profile_cfg"]["profile_version"]
        if (
            profile_version["major"] != 9
            or profile_version["minor"] != 1
            or profile_version["patch"] != 0
        ):
            raise KeyError(
                f"ERROR {profile_json} because 'profile_version' is not "
                + f"supported: {profile_version}"
            )

    if use_summary:
        with open(summary_json, "r") as f:
            summary_data = json.load(f)

        summary_file = summary_json

    iduc = os.path.basename(profile_json)
    iduc = iduc.replace(".json", "")

    df_row = {
        "id": iduc,
        "profile_name": None,
        "device_clock_Hz": None,
        "core_clock_Hz": None,
        "common_lane_rate_Hz": None,
        "rx_jesd_mode": None,
        "tx_jesd_mode": None,
        # "failed_reason": None,
        "is_8t8r": None,
        "jesd_settings": None,
        "datapath": None,
        # "jif_model": None,
        # "dts_file": None,
        # "bin_filename": None,
        # "hdl_build_id": None,
    }

    if use_summary:
        if "is_8t8r" not in summary_data:
            raise Exception("AD9088 is not supported")

        if "is_8t8r" in summary_data and summary_data["is_8t8r"]:
            raise Exception("AD9088 is not supported")

    else:
        df_row["is_8t8r"] = profile_data["profile_cfg"]["is_8t8r"]
        if df_row["is_8t8r"]:
            raise Exception("AD9088 is not supported")

    if use_summary:
        device_clock_Hz = summary_data["general_info"]["device_clock_Hz"]
    else:
        device_clock_Hz = profile_data["clk_cfg"]["dev_clk_freq_kHz"] * 1000
    if device_clock_Hz is None:
        raise KeyError(
            f"Skipping {profile_data} because 'device_clock_Hz' key is missing"
        )
    df_row["device_clock_Hz"] = device_clock_Hz

    if use_summary:
        core_clock_Hz = summary_data["general_info"]["fpga_clock_Hz"]
        if core_clock_Hz is None:
            raise KeyError(
                f"Skipping {profile_data} because 'core_clock_Hz' key is missing"
            )
        df_row["core_clock_Hz"] = core_clock_Hz

    if use_summary:
        common_lane_rate_Hz = summary_data["general_info"]["common_lane_rate_Hz"]
        if common_lane_rate_Hz is None:
            raise KeyError(
                f"Skipping {profile_data} because 'common_lane_rate_Hz' key is missing"
            )
    else:
        common_lane_rate_Hz = (
            profile_data["jtx"][0]["common_link_cfg"]["lane_rate_kHz"] * 1000
        )

    df_row["common_lane_rate_Hz"] = common_lane_rate_Hz

    # Parse datapath config
    path = 0
    if use_summary:
        cddc_decimation = summary_data["rx_routes"][path]["cdrc"]
        fddc_decimation = summary_data["rx_routes"][path]["fdrc"]
        cduc_interpolation = summary_data["tx_routes"][path]["cdrc"]
        fduc_interpolation = summary_data["tx_routes"][path]["fdrc"]
    else:
        cddc_decimation = profile_data["rx_path"][0]["rx_cddc"][0]["drc_ratio"]
        fddc_decimation = profile_data["rx_path"][0]["rx_fddc"][0]["drc_ratio"]
        cduc_interpolation = profile_data["tx_path"][0]["tx_cduc"][0]["drc_ratio"]
        fduc_interpolation = profile_data["tx_path"][0]["tx_fduc"][0]["drc_ratio"]

        if cddc_decimation == 0:
            cddc_decimation = 1
        elif cddc_decimation == 1:
            cddc_decimation = 2
        elif cddc_decimation == 2:
            cddc_decimation = 3
        elif cddc_decimation == 3:
            cddc_decimation = 4
        elif cddc_decimation == 4:
            cddc_decimation = 6
        elif cddc_decimation == 5:
            cddc_decimation = 12

        fddc_decimation = int(fddc_decimation)
        fddc_decimation = 2**fddc_decimation

        cduc_interpolation = int(cduc_interpolation)

        # THIS IS REALLY BIZARRE but how the profile gen works
        fduc_interpolation = int(fduc_interpolation)

    df_row["datapath"] = {
        "cddc_decimation": cddc_decimation,
        "fddc_decimation": fddc_decimation,
        "cduc_interpolation": cduc_interpolation,
        "fduc_interpolation": fduc_interpolation,
    }

    # Parse JESD204 config
    link_index = 0
    lane_index = 0
    rx_jesd_mode = profile_data["jtx"][link_index]["tx_link_cfg"][lane_index][
        "quick_mode_id"
    ]
    tx_jesd_mode = profile_data["jrx"][link_index]["rx_link_cfg"][lane_index][
        "quick_mode_id"
    ]

    if profile_data["jtx"][link_index]["common_link_cfg"]["ver"] == 0:
        # JESD204B
        raise Exception(f"Skipping {summary_file} because it is a JESD204B profile")
    if profile_data["jrx"][link_index]["common_link_cfg"]["ver"] == 0:
        # JESD204B
        raise Exception(f"Skipping {summary_file} because it is a JESD204B profile")

    df_row["jesd_settings"] = {}
    for rtx, cfg in zip(["jtx", "jrx"], ["tx_link_cfg", "rx_link_cfg"]):
        df_row["jesd_settings"][rtx] = {}
        for setting, jesd_setting_key in zip(
            ["L", "F", "M", "S", "HD", "K", "N", "Np"],
            [
                "l_minus1",
                "f_minus1",
                "m_minus1",
                "s_minus1",
                "high_dens",
                "k_minus1",
                "n_minus1",
                "np_minus1",
            ],
        ):
            if setting in ["N", "K"]:
                continue
            if jesd_setting_key not in profile_data[rtx][link_index][cfg][lane_index]:
                raise KeyError(
                    f"Skipping {summary_file} because {jesd_setting_key} "
                    + f"key is missing in {rtx} {cfg}"
                )
            if setting == "HD":
                df_row["jesd_settings"][rtx][setting] = profile_data[rtx][link_index][
                    cfg
                ][lane_index][jesd_setting_key]
            else:
                df_row["jesd_settings"][rtx][setting] = (
                    int(
                        profile_data[rtx][link_index][cfg][lane_index][jesd_setting_key]
                    )
                    + 1
                )

    if rx_jesd_mode is None:
        raise KeyError(f"Skipping {summary_file} because 'rx_jesd_mode' key is missing")
    df_row["rx_jesd_mode"] = rx_jesd_mode
    if tx_jesd_mode is None:
        raise KeyError(f"Skipping {summary_file} because 'tx_jesd_mode' key is missing")
    df_row["tx_jesd_mode"] = tx_jesd_mode
    profile_name = os.path.basename(full_profile_filename)
    df_row["profile_name"] = profile_name.replace(".json", "")

    return df_row


def apply_settings(conv: converter, profile_settings: Dict) -> None:
    """Apply settings to the AD9084 converter.

    Args:
        conv (converter): The AD9084 converter object.
        profile_settings (Dict): The profile settings dictionary
            containing configuration data.

    Raises:
        ValueError: If the TX and RX JESD204C modes do not match.
    """
    jesd_class = "jesd204c"  # only JESD204C is supported for AD9084 right now

    cddc_dec = profile_settings["datapath"]["cddc_decimation"]
    cddc_dec = int(cddc_dec)
    fddc_dec = profile_settings["datapath"]["fddc_decimation"]
    fddc_dec = int(fddc_dec)
    converter_rate = int(profile_settings["device_clock_Hz"])

    conv.sample_clock = converter_rate / (cddc_dec * fddc_dec)
    conv.datapath.cddc_decimations = [cddc_dec] * 4
    conv.datapath.fddc_decimations = [fddc_dec] * 8
    conv.datapath.fddc_enabled = [True] * 8

    M = profile_settings["jesd_settings"]["jtx"]["M"]
    L = profile_settings["jesd_settings"]["jtx"]["L"]
    S = profile_settings["jesd_settings"]["jtx"]["S"]
    Np = profile_settings["jesd_settings"]["jtx"]["Np"]

    # Verify TX and RX JESD204C modes are the same
    M_tx = profile_settings["jesd_settings"]["jrx"]["M"]
    L_tx = profile_settings["jesd_settings"]["jrx"]["L"]
    S_tx = profile_settings["jesd_settings"]["jrx"]["S"]
    Np_tx = profile_settings["jesd_settings"]["jrx"]["Np"]
    if M != M_tx or L != L_tx or S != S_tx or Np != Np_tx:
        raise ValueError(
            f"TX and RX JESD204C modes do not match: "
            f"TX (M={M_tx}, L={L_tx}, S={S_tx}, Np={Np_tx}), "
            f"RX (M={M}, L={L}, S={S}, Np={Np})"
        )

    mode_rx = get_jesd_mode_from_params(
        conv, M=M, L=L, S=S, Np=Np, jesd_class=jesd_class
    )
    assert mode_rx, f"Could not find JESD204C mode for M={M}, L={L}, S={S}, Np={Np}"
    if isinstance(mode_rx, list) and len(mode_rx) > 1:
        print(
            "WARNING: Multiple JESD204C modes found for "
            + f"M={M}, L={L}, S={S}, Np={Np}\n"
            + "Picking the first one."
        )
    mode_rx = mode_rx[0]["mode"]

    conv.set_quick_configuration_mode(mode_rx, jesd_class)


if __name__ == "__main__":
    data = _load_rx_config_modes()
    # _load_tx_config_modes()



# flake8: noqa
import pprint

import pytest

import adijif


@pytest.mark.parametrize("part", ["ad9081_rx", "ad9082_rx"])
def test_ad9081_core_rx_solver(part):
    vcxo = 100000000

    sys = adijif.system(part, "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.Debug_Solver = False
    # sys.fpga.request_device_clock = False
    sys.fpga.force_qpll = True

    sys.converter.clocking_option = "integrated_pll"

    # Get Converter clocking requirements
    sys.converter.sample_clock = 250e6
    # sys.converter.decimation = 16
    sys.converter.datapath.cddc_decimations = [4, 4, 4, 4]
    sys.converter.datapath.fddc_decimations = [4, 4, 4, 4, 4, 4, 4, 4]
    sys.converter.datapath.fddc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    assert sys.converter.decimation == 16
    sys.converter.L = 4
    sys.converter.M = 8
    sys.converter.N = 16
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 4
    # sys.converter.S = 1

    # sys._try_fpga_configs()
    cfg = sys.solve()

    # assert sys.fpga.configs[0]["qpll_0_cpll_1"].value[0] == 0  # QPLL
    assert cfg["fpga_AD9081_RX"]["type"] == "qpll"


@pytest.mark.parametrize("part", ["ad9081_tx", "ad9082_tx"])
def test_ad9081_core_tx_solver(part):
    vcxo = 100000000

    sys = adijif.system(part, "hmc7044", "xilinx", vcxo)
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.Debug_Solver = False
    # sys.fpga.request_device_clock = False

    sys.converter.clocking_option = "integrated_pll"

    # Get Converter clocking requirements
    sys.converter.sample_clock = 250e6
    # sys.converter.interpolation = 48
    sys.converter.datapath.cduc_interpolation = 12
    sys.converter.datapath.fduc_interpolation = 4
    sys.converter.datapath.fduc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    assert sys.converter.interpolation == 48
    mode = adijif.utils.get_jesd_mode_from_params(
        sys.converter, L=4, M=8, Np=16, K=32, F=4
    )
    assert len(mode) == 1
    print(mode)
    jmode = mode[0]["mode"]
    jclass = mode[0]["jesd_class"]
    sys.converter.set_quick_configuration_mode(jmode, jclass)
    assert sys.converter.L == 4
    assert sys.converter.M == 8
    assert sys.converter.Np == 16
    assert sys.converter.K == 32
    assert sys.converter.F == 4
    assert sys.converter.S == 1

    # sys._try_fpga_configs()
    cfg = sys.solve()
    # pprint.pprint(cfg)

    assert cfg["fpga_AD9081_TX"]["type"] == "qpll"


@pytest.mark.parametrize("part", ["ad9081", "ad9082"])
def test_ad9081_core_rxtx_solver(part):
    vcxo = 100000000

    sys = adijif.system(part, "hmc7044", "xilinx", vcxo, solver="CPLEX")
    # sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo)
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.Debug_Solver = True
    # sys.fpga.request_device_clock = False
    sys.fpga.force_qpll = True

    sys.converter.dac.clocking_option = "integrated_pll"
    sys.converter.adc.clocking_option = "integrated_pll"
    sys.converter.clocking_option = "integrated_pll"

    # Get Converter clocking requirements
    sys.converter.dac.jesd_class = "jesd204b"
    sys.converter.dac.sample_clock = 250e6
    # sys.converter.dac.interpolation = 48
    sys.converter.dac.datapath.cduc_interpolation = 12
    sys.converter.dac.datapath.fduc_interpolation = 4
    sys.converter.dac.datapath.fduc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    sys.converter.dac.L = 4
    sys.converter.dac.M = 8
    sys.converter.dac.N = 16
    sys.converter.dac.Np = 16
    sys.converter.dac.K = 32
    sys.converter.dac.F = 4
    print("S DAC:", sys.converter.dac.S)
    # sys.converter.S = 1

    # Get Converter clocking requirements
    sys.converter.adc.jesd_class = "jesd204b"
    sys.converter.adc.sample_clock = 250e6
    # sys.converter.adc.decimation = 16
    sys.converter.adc.datapath.cddc_decimations = [4, 4, 4, 4]
    sys.converter.adc.datapath.fddc_decimations = [4, 4, 4, 4, 4, 4, 4, 4]
    sys.converter.adc.datapath.fddc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    sys.converter.adc.L = 4
    sys.converter.adc.M = 8
    sys.converter.adc.N = 16
    sys.converter.adc.Np = 16
    sys.converter.adc.K = 32
    sys.converter.adc.F = 4
    sys.converter.adc.HD = 0
    sys.converter.adc.HD = 0

    print("S ADC:", sys.converter.adc.S)
    # sys.converter.S = 1

    # sys._try_fpga_configs()
    o = sys.solve()

    # print(sys.solution)
    # print(dir(sys.solution))
    # sys.solution.print_solution()

    # pprint.pprint(o)

    assert o["fpga_adc"]["type"] == "qpll"
    assert o["fpga_dac"]["type"] == "qpll"


def test_ad9081_rxtx_zcu102_default_config():
    vcxo = 100e6

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.Debug_Solver = False
    sys.converter.clocking_option = "integrated_pll"
    sys.fpga.out_clk_select = "XCVR_REFCLK"  # force reference to be core clock rate
    sys.converter.adc.sample_clock = 4000000000 // (4 * 4)
    sys.converter.dac.sample_clock = 12000000000 // (8 * 6)

    # sys.converter.adc.decimation = 16
    # sys.converter.dac.interpolation = 48
    sys.converter.adc.datapath.cddc_decimations = [4, 4, 4, 4]
    sys.converter.dac.datapath.cduc_interpolation = 8
    sys.converter.adc.datapath.fddc_decimations = [4, 4, 4, 4, 4, 4, 4, 4]
    sys.converter.adc.datapath.fddc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    sys.converter.dac.datapath.fduc_interpolation = 6
    sys.converter.dac.datapath.fduc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]

    assert sys.converter.adc.decimation == 16
    assert sys.converter.dac.interpolation == 48

    mode_tx = "9"
    mode_rx = "10.0"

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")

    assert sys.converter.adc.M == 8
    assert sys.converter.adc.F == 4
    assert sys.converter.adc.K == 32
    assert sys.converter.adc.Np == 16
    assert sys.converter.adc.CS == 0
    assert sys.converter.adc.L == 4
    assert sys.converter.adc.S == 1
    # assert sys.converter.adc.HD == 1

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    cfg = sys.solve()

    print("Mode passed: ", mode_tx, sys.converter.adc.decimation)


def test_ad9081_rxtx_zcu102_lowrate_config():
    vcxo = 100e6

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zcu102")
    # sys.fpga.sys_clk_select = "GTH34_SYSCLK_QPLL0"  # Use faster QPLL
    sys.Debug_Solver = False
    sys.converter.clocking_option = "integrated_pll"
    # sys.fpga.out_clk_select = "XCVR_REFCLK"  # force reference to be core clock rate
    sys.converter.adc.sample_clock = 4000000000 / (4 * 8)
    sys.converter.dac.sample_clock = 4000000000 / (4 * 8)

    # sys.converter.adc.decimation = 4 * 8
    # sys.converter.dac.interpolation = 4 * 8
    sys.converter.adc.datapath.cddc_decimations = [4, 4, 4, 4]
    sys.converter.dac.datapath.cduc_interpolation = 4
    sys.converter.adc.datapath.fddc_decimations = [8, 8, 8, 8, 8, 8, 8, 8]
    sys.converter.dac.datapath.fduc_interpolation = 8
    sys.converter.adc.datapath.fddc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    sys.converter.dac.datapath.fduc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]

    mode_tx = "5"
    mode_rx = "6.0"

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")
    # sys.converter._skip_clock_validation = True  # slightly too slow for low rate
    # sys.converter._skip_clock_validation = True  # slightly too slow for low rate

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    cfg = sys.solve()

    assert cfg["jesd_dac"]["bit_clock"] == 5e9
    assert cfg["jesd_adc"]["bit_clock"] == 5e9

    # cfg["fpga_dac"]["d"] = 1
    # cfg["fpga_adc"]["d"] = 2

    # pprint.pprint(cfg)

    print("Mode passed: ", mode_tx, sys.converter.adc.decimation)


def test_ad9081_np16_verify_no_extra_link_clock():
    vcxo = 100e6

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zcu102")
    # sys.fpga.sys_clk_select = "GTH34_SYSCLK_QPLL0"  # Use faster QPLL
    sys.Debug_Solver = False
    sys.converter.clocking_option = "integrated_pll"
    # sys.fpga.out_clk_select = "XCVR_REFCLK"  # force reference to be core clock rate
    sys.converter.adc.sample_clock = 4000000000 / (4 * 8)
    sys.converter.dac.sample_clock = 4000000000 / (4 * 8)

    # sys.converter.adc.decimation = 4 * 8
    # sys.converter.dac.interpolation = 4 * 8
    sys.converter.adc.datapath.cddc_decimations = [4, 4, 4, 4]
    sys.converter.dac.datapath.cduc_interpolation = 4
    sys.converter.adc.datapath.fddc_decimations = [8, 8, 8, 8, 8, 8, 8, 8]
    sys.converter.dac.datapath.fduc_interpolation = 8
    sys.converter.adc.datapath.fddc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    sys.converter.dac.datapath.fduc_enabled = [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]

    mode_tx = "5"
    mode_rx = "6.0"

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")
    assert sys.converter.adc.Np == 16
    assert sys.converter.dac.Np == 16

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    cfg = sys.solve()

    assert cfg["jesd_dac"]["bit_clock"] == 5e9
    assert cfg["jesd_adc"]["bit_clock"] == 5e9

    # cfg["fpga_dac"]["d"] = 1
    # cfg["fpga_adc"]["d"] = 2

    # pprint.pprint(cfg)

    assert cfg["fpga_dac"]["device_clock_source"] == "external"
    assert cfg["fpga_adc"]["device_clock_source"] == "external"


def test_ad9081_np12_verify_extra_link_clock():
    # Reference https://github.com/analogdevicesinc/linux/blob/master/arch/microblaze/boot/dts/vcu118_quad_ad9081_204c_txmode_23_rxmode_25_revc.dts
    vcxo = 100e6

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("vcu118")
    sys.Debug_Solver = False
    sys.converter.clocking_option = "integrated_pll"
    sys.converter.adc.sample_clock = 4000000000 / (2 * 1)
    sys.converter.dac.sample_clock = 12000000000 / (6 * 1)
    sys.fpga.out_clk_select = (
        "XCVR_PROGDIV_CLK"  # force reference to be core clock rate
    )

    # sys.converter.adc.decimation = 2 * 1
    # sys.converter.dac.interpolation = 6 * 1
    sys.converter.adc.datapath.cddc_decimations = [2, 2, 2, 2]
    sys.converter.dac.datapath.cduc_interpolation = 6
    sys.converter.adc.datapath.fddc_decimations = [1, 1, 1, 1, 1, 1, 1, 1]
    sys.converter.dac.datapath.fduc_interpolation = 1

    mode_tx = "23"
    mode_rx = "25.0"

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204c")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204c")

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    assert sys.converter.dac.M == 4
    assert sys.converter.dac.F == 3
    assert sys.converter.dac.K == 256
    assert sys.converter.dac.Np == 12
    assert sys.converter.dac.CS == 0
    assert sys.converter.dac.L == 4
    assert sys.converter.dac.S == 2
    assert sys.converter.dac.HD == 0

    assert sys.converter.adc.M == 4
    assert sys.converter.adc.F == 3
    assert sys.converter.adc.K == 256
    assert sys.converter.adc.Np == 12
    assert sys.converter.adc.CS == 0
    assert sys.converter.adc.L == 4
    assert sys.converter.adc.S == 2
    assert sys.converter.adc.HD == 0

    cfg = sys.solve()
    pprint.pprint(cfg)

    assert cfg["fpga_dac"]["device_clock_source"] == "external"
    assert cfg["fpga_adc"]["device_clock_source"] == "external"

    assert cfg["jesd_dac"]["bit_clock"] == 24750000000
    assert cfg["jesd_adc"]["bit_clock"] == 24750000000

    assert cfg["fpga_adc"]["out_clk_select"] == "XCVR_PROGDIV_CLK"

    if cfg["fpga_adc"]["sys_clk_select"] in ["CPLL", "QPLL", "QPLL0", "QPLL1"]:
        in_clock_rate = cfg["fpga_adc"]["bit_clock"]
    elif cfg["fpga_adc"]["sys_clk_select"] == "XCVR_PROGDIV_CLK":
        in_clock_rate = cfg["fpga_adc"]["bit_clock"] / 66
    assert (
        cfg["jesd_adc"]["bit_clock"] / cfg["fpga_adc"]["progdiv"]
        == cfg["jesd_adc"]["bit_clock"] / 66
    )

    assert (
        cfg["fpga_adc"]["transport_samples_per_clock"]
        * cfg["clock"]["output_clocks"]["vcu118_adc_device_clk"]["rate"]
        == sys.converter.adc.sample_clock
    )


def test_ad9081_zc706_gtx_gap():
    # Determine AD9081+ZCU102 Configuration For RX and TX contrained together

    import pprint

    import adijif

    vcxo = 100e6

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.ref_clock_constraint = "Unconstrained"
    # sys.fpga.sys_clk_select = "GTH34_SYSCLK_QPLL0"  # Use faster QPLL
    # sys.fpga.sys_clk_select = "CPLL"  # Use faster QPLL

    sys.converter.clocking_option = "integrated_pll"
    # sys.fpga.out_clk_select = "XCVR_PROGDIV_CLK"  # force reference to be core clock rate
    # sys.converter.adc.sample_clock = 4000000000 / (4 * 4)
    # sys.converter.dac.sample_clock = 12000000000 / (8 * 6)
    sys.converter.adc.sample_clock = 4000000000 * (232 / 250) / (4 * 4)
    sys.converter.dac.sample_clock = 12000000000 * (232 / 250) / (8 * 6)
    # sys.converter.adc.sample_clock = 4000000000 * (0.1) / (4 * 4)
    # sys.converter.dac.sample_clock = 12000000000 * (0.1) / (8 * 6)

    sys.converter.adc.datapath.cddc_decimations = [4] * 4
    sys.converter.adc.datapath.fddc_decimations = [4] * 8
    sys.converter.adc.datapath.fddc_enabled = [True] * 8
    sys.converter.dac.datapath.cduc_interpolation = 8
    sys.converter.dac.datapath.fduc_interpolation = 6
    sys.converter.dac.datapath.fduc_enabled = [True] * 8

    mode_tx = "9"
    mode_rx = "10.0"  # why are these requiredmM, is it possible to find the mode based on JSED parameters?

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")

    """
    assert sys.converter.adc.M == 8
    assert sys.converter.adc.F == 4
    assert sys.converter.adc.K == 32
    assert sys.converter.adc.Np == 16
    assert sys.converter.adc.CS == 0
    assert sys.converter.adc.L == 4
    assert sys.converter.adc.S == 1
    """
    print(sys.converter.adc.bit_clock)
    print(sys.converter.dac.bit_clock)

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    with pytest.raises(Exception, match=".*No solution found.*"):
        cfg = sys.solve()



"""Utility functions and base classes for the explorer."""

from abc import ABC, abstractmethod

import streamlit as st

# st.set_page_config(layout="wide")


class Page(ABC):
    """Base class for all page types."""

    @abstractmethod
    def write(self) -> None:
        """Render the page content."""
        pass


def add_custom_css() -> None:
    """Add custom CSS styling to the Streamlit app."""
    st.html(
        """
        <style>
            .stMainBlockContainer {
                max-width:1200px;
            }
        </style>
        """,
    )
    # st .markdown(
    #     """
    #     <style>
    #         section.main > div {max-width:1000px;}
    #     </style>
    #     """,
    #     unsafe_allow_html=True,
    # )



"""Drawing features for AD9084."""

from typing import Dict

from adijif.draw import Layout, Node  # type: ignore # isort: skip  # noqa: I202


class ad9084_draw:
    """AD9084 drawing features."""

    _system_draw = False
    show_rates = True

    def _init_diagram(self) -> None:
        """Initialize diagram for AD9084 alone."""
        self.ic_diagram_node = None
        self._diagram_output_dividers = []

        name = "AD9084" if "9084" in self.name else "AD9088"
        N = 4 if "9084" in self.name else 8

        self.ic_diagram_node = Node(name)

        # External
        # ref_in = Node("REF_IN", ntype="input")
        # lo.add_node(ref_in)

        crossbar = Node("MUX0", ntype="crossbar")
        crossbar_rm = Node("Router MUX", ntype="crossbar")

        self.ic_diagram_node.add_child(crossbar)
        self.ic_diagram_node.add_child(crossbar_rm)

        for adc in range(N):
            adc_node = Node(f"ADC{adc}", ntype="adc")
            self.ic_diagram_node.add_child(adc_node)
            adc_node.shape = "parallelogram"
            self.ic_diagram_node.add_connection({"from": adc_node, "to": crossbar})

        for cddc in range(N):
            cddc_node = Node(f"CDDC{cddc}", ntype="ddc")
            self.ic_diagram_node.add_child(cddc_node)
            self.ic_diagram_node.add_connection({"from": crossbar, "to": cddc_node})
            self.ic_diagram_node.add_connection({"from": cddc_node, "to": crossbar_rm})

        for fddc in range(N * 2):
            fddc_node = Node(f"FDDC{fddc}", ntype="ddc")
            self.ic_diagram_node.add_child(fddc_node)
            self.ic_diagram_node.add_connection({"from": crossbar_rm, "to": fddc_node})

        jesd204_framer = Node("JESD204 Framer", ntype="jesd204framer")
        self.ic_diagram_node.add_child(jesd204_framer)

        for ddc in range(N * 2):
            fddc = self.ic_diagram_node.get_child(f"FDDC{ddc}")
            self.ic_diagram_node.add_connection({"from": fddc, "to": jesd204_framer})

    def _update_diagram(self, config: Dict) -> None:
        """Update diagram with configuration.

        Args:
            config (Dict): Configuration dictionary

        Raises:
            Exception: If key is not D followed by a number
        """
        # Add output dividers
        keys = config.keys()
        output_dividers = self.ic_diagram_node.get_child("Output Dividers")
        for key in keys:
            if key.startswith("D"):
                div = Node(key, ntype="divider")
                output_dividers.add_child(div)
                self.ic_diagram_node.add_connection(
                    {"from": output_dividers, "to": div}
                )
            else:
                raise Exception(
                    f"Unknown key {key}. Must be of for DX where X is a number"
                )

    def draw(
        self, clocks: Dict, lo: Layout = None, clock_chip_node: Node = None
    ) -> str:
        """Draw diagram in d2 language for IC alone with reference clock.

        Args:
            clocks (Dict): Dictionary of clocks
            lo (Layout): Layout object to add to. Defaults to None.
            clock_chip_node (Node): Node to connect to. Defaults to None.

        Returns:
            str: Diagram in d2 language

        Raises:
            Exception: If no solution is saved
        """
        if not self._saved_solution:
            raise Exception("No solution to draw. Must call solve first.")

        system_draw = lo is not None

        name = "AD9084" if "9084" in self.name else "AD9088"
        N = 4 if "9084" in self.name else 8

        if not system_draw:
            lo = Layout(f"{name} Example")
            lo.show_rates = self.show_rates
        else:
            # Verify lo is a Layout object
            assert isinstance(lo, Layout), "lo must be a Layout object"
        lo.add_node(self.ic_diagram_node)

        if not system_draw:
            ref_in = Node("REF_IN", ntype="input")
            lo.add_node(ref_in)
        else:
            to_node = lo.get_node(f"{name}_ref_clk")
            # Locate node connected to this one
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            ref_in = from_node[0]["from"]
            # Remove to_node since it is not needed
            lo.remove_node(to_node.name)

        for i in range(N):
            adc = self.ic_diagram_node.get_child(f"ADC{i}")
            lo.add_connection(
                {"from": ref_in, "to": adc, "rate": clocks[f"{name}_ref_clk"]}
            )

        # Update Node values
        fddc_index = 0
        for cddc in range(N):
            rate = clocks[f"{name}_ref_clk"]
            self.ic_diagram_node.update_connection("MUX0", f"CDDC{cddc}", rate)

            cddc_node = self.ic_diagram_node.get_child(f"CDDC{cddc}")
            cddc_node.value = str(self.datapath.cddc_decimations[cddc])
            drate = rate / self.datapath.cddc_decimations[cddc]

            self.ic_diagram_node.update_connection(f"CDDC{cddc}", "Router MUX", drate)

            self.ic_diagram_node.update_connection(
                "Router MUX", f"FDDC{fddc_index}", drate
            )
            fddc_rate_out = drate / self.datapath.fddc_decimations[fddc_index]
            self.ic_diagram_node.update_connection(
                f"FDDC{fddc_index}", "JESD204 Framer", fddc_rate_out
            )
            fddc_index += 1

            self.ic_diagram_node.update_connection(
                "Router MUX", f"FDDC{fddc_index}", drate
            )
            fddc_rate_out = drate / self.datapath.fddc_decimations[fddc_index]
            self.ic_diagram_node.update_connection(
                f"FDDC{fddc_index}", "JESD204 Framer", fddc_rate_out
            )
            fddc_index += 1

        # Connect clock to framer
        if not system_draw:
            sysref_in = Node("SYSREF_IN", ntype="input")

            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": self.ic_diagram_node.get_child("JESD204 Framer"),
                    "rate": clocks[f"{name}_sysref"],
                }
            )
        else:
            to_node = lo.get_node(f"{name}_sysref")
            # Locate node connected to this one
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            sysref_in = from_node[0]["from"]
            # Remove to_node since it is not needed
            lo.remove_node(to_node.name)

            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": self.ic_diagram_node.get_child("JESD204 Framer"),
                    "rate": clocks[f"{name}_sysref"],
                }
            )

        # Connect Remote Deframer
        remote_deframer = Node("JESD204 Deframer", ntype="deframer")

        # Add connect for each lane
        for _ in range(self.L):
            lane_rate = self.bit_clock
            lo.add_connection(
                {
                    "from": self.ic_diagram_node.get_child("JESD204 Framer"),
                    "to": remote_deframer,
                    "rate": lane_rate,
                }
            )

        if not system_draw:
            return lo.draw()



"""AD9081 MxFE Utility Functions."""

import csv
import os
from typing import Dict, Union


def _convert_to_config(
    mode: str,
    L: Union[int, float],
    M: Union[int, float],
    F: Union[int, float],
    S: Union[int, float],
    HD: Union[int, float],
    K: Union[int, float],
    N: Union[int, float],
    Np: Union[int, float],
    CS: Union[int, float],
    E: Union[int, float],
    global_index: Union[int, float],
    jesd_class: str,
) -> Dict:
    return {
        "L": L,
        "M": M,
        "F": F,
        "S": S,
        # "HD": 1 if F == 1 else 0,
        "Np": Np,
        "K": K,
        "HD": HD,
        "CS": CS,
        "E": E,
        "jesd_class": jesd_class,
        "global_index": global_index,
    }


def _load_rx_config_modes() -> Dict:
    """Load RX JESD configuration tables from file."""
    return {
        "jesd204b": _read_table("ad9081_JTx_204B.csv", True),
        "jesd204c": _read_table("ad9081_JTx_204C.csv", False),
    }


def _load_tx_config_modes() -> Dict:
    """Load TX JESD configuration tables from file."""
    return {
        "jesd204b": _read_table("ad9081_JRx_204B.csv", True),
        "jesd204c": _read_table("ad9081_JRx_204C.csv", False),
    }


def _read_table(fn: str, jesd204b: bool, qcm: Union[Dict, None] = None) -> Dict:
    loc = os.path.dirname(__file__)
    fn = os.path.join(loc, "resources", fn)
    with open(fn) as f:
        records = csv.DictReader(f)
        if qcm:
            quick_configuration_modes = qcm
        else:
            quick_configuration_modes = {}
        for row in records:
            d = dict(row)
            if "_JTx_" in fn:
                HD = int(float(d["HD"]))
                conv_min = float(d["FADC Min (GSPS)"])
                conv_max = float(d["FADC Max (GSPS)"])
                lane_min = float(d["Lane Rate Min (GSPS)"])
                lane_max = float(d["Lane Rate Max (GSPS)"])
            else:
                HD = 0
                conv_min = -1
                conv_max = -1
                lane_min = -1
                lane_max = -1
            if jesd204b:
                k = "JESD204B Mode Number"
                jmode = "jesd204b"
                E = 0
            else:
                k = "JESD204C Mode Number"
                jmode = "jesd204c"
                E = int(float(d["E"]))
            mode = d[k]
            del d[k]
            if not mode:
                continue
            gi = int(float(d["global_index"]))
            decs = {
                "coarse": int(float(d["Coarse"])),
                "fine": int(float(d["Fine"])),
                "conv_min": conv_min,
                "conv_max": conv_max,
                "lane_min": lane_min,
                "lane_max": lane_max,
            }
            cfg = _convert_to_config(
                mode=mode,
                L=int(float(d["L"])),
                M=int(float(d["M"])),
                F=int(float(d["F"])),
                S=int(float(d["S"])),
                N=int(float(d["NP"])),
                Np=int(float(d["NP"])),
                CS=int(0),
                HD=HD,
                K=int(float(d["K"])),
                E=E,
                global_index=gi,
                jesd_class=jmode,
            )
            if mode not in quick_configuration_modes:
                quick_configuration_modes[mode] = cfg
                quick_configuration_modes[mode]["decimations"] = [decs]
            else:
                quick_configuration_modes[mode]["decimations"].append(decs)
    return quick_configuration_modes


if __name__ == "__main__":
    _load_rx_config_modes()
    _load_tx_config_modes()



"""Clock configurator page for PyADI-JIF Tools Explorer."""

import logging
from typing import Optional

import streamlit as st

from adijif.clocks import supported_parts as sp

from ..utils import Page

try:
    import adidt as dt
except ImportError:
    dt = None

log = logging.getLogger(__name__)
log.setLevel(logging.ERROR)

# Parse parts and options to skip
options_to_skip = ["list_references_available", "d_syspulse"]
parts_to_ignore = ["ad9545", "ad9523_1"]
sp = [p for p in sp if p not in parts_to_ignore]
sp = [p for p in sp if p != "hmc7044"]
sp.insert(0, "hmc7044")


class ClockConfigurator(Page):
    """Clock configurator tool page."""

    def __init__(self, state: Optional[object]) -> None:
        """Initialize clock configurator page.

        Args:
            state: Application state object
        """
        self.state = state

    def write(self) -> None:
        """Render the clock configurator page."""
        st.title("Clock Configurator")
        # sp = ["hmc7044"]

        sb = st.selectbox(
            label="Select a part",
            options=sp,
            format_func=lambda x: x.upper().replace("_", "-"),
        )

        with st.expander("Clock Inputs and Outputs", expanded=True):
            reference = st.number_input(
                "Reference Clock",
                value=125000000,
                min_value=1,
                max_value=int(1e9),
                step=1,
            )

            with st.container(border=True):
                num_outputs = st.number_input(
                    "Number of Clock Outputs",
                    value=2,
                    min_value=1,
                    max_value=10,
                    step=1,
                )
                outputs = []
                output_names = []
                for i in range(num_outputs):
                    columns = st.columns(2)
                    with columns[0]:
                        outputs.append(
                            st.number_input(
                                f"Output Clock {i+1}",
                                value=125000000,
                                min_value=1,
                                max_value=int(1e9),
                                step=1,
                            )
                        )
                    with columns[1]:
                        output_names.append(
                            st.text_input(f"Output Clock Name {i+1}", f"CLK{i+1}")
                        )

        import adijif  # noqa: F401

        class_def = eval(f"adijif.{sb}")  # noqa: S307
        props = dir(class_def)

        props = [p for p in props if not p.startswith("__")]
        props = [p for p in props if not p.startswith("_")]
        configurable_props = [
            p.replace("_available", "") for p in props if "_available" in p
        ]

        prop_and_options = {}
        for prop in configurable_props:
            if prop in options_to_skip or prop + "_available" not in props:
                continue
            prop_and_options[prop] = getattr(class_def, prop + "_available")

        selections = {}
        with st.expander("Internal Clock Configuration"):
            for prop, options in prop_and_options.items():
                prop_docstring = getattr(class_def, prop).__doc__
                with st.container(border=True):
                    label = f"""
                    {prop} : {prop_docstring}"""
                    if len(options) > 16:
                        v_min, v_max = min(options), max(options)
                        start, end = st.select_slider(
                            label, options, value=(v_min, v_max)
                        )
                        selections[prop] = {"start": start, "end": end}
                    else:
                        selections[prop] = st.multiselect(label, options)

        clk_chip = eval(f"adijif.{sb}()")  # noqa: S307

        for prop, values in selections.items():
            if isinstance(values, dict):
                props_available = getattr(clk_chip, prop + "_available")
                if (
                    min(props_available) == values["start"]
                    and max(props_available) == values["end"]
                ):
                    continue
                picked = [
                    v for v in props_available if values["start"] <= v <= values["end"]
                ]
                setattr(clk_chip, prop, picked)
            else:
                if not values:
                    continue
                setattr(clk_chip, prop, values)

        output_clocks = outputs
        output_clocks = list(map(int, output_clocks))  # force to be ints

        # Remove duplicates
        output_clocks_filtered = output_clocks
        output_names_filtered = output_names

        clk_chip.set_requested_clocks(
            reference, output_clocks_filtered, output_names_filtered
        )

        try:
            clk_chip.solve()
            o = clk_chip.get_config()
            log.info("Solution found")
            log.info(o)

            # if dt:
            #     columns_results = st.columns(2)

            #     with columns_results[0]:
            #         with st.expander("Found Configuration"):
            #             st.write(o)

            #     with columns_results[1]:
            #         if sb == "hmc7044":

            #             clk = dt.hmc7044_dt(offline=True)
            #             dtsi = clk.map_config_to_fragment(o)
            #             with st.expander("Device Tree Fragment"):
            #                 st.code(dtsi)
            # else:
            #     with st.expander("Found Configuration"):
            #         st.write(o)

            config_out = o
            image_data = clk_chip.draw()

            warning = False

        except Exception as e:
            log.warning(e)

            warning = True

        with st.expander("Found Configuration", expanded=True):
            if warning:
                st.warning("No valid configuration found")
            else:
                st.write(config_out)

        with st.expander("Diagram", expanded=True):
            if warning:
                st.warning("No diagram to show")
            else:
                st.image(image_data, width="stretch")



"""Console script for adijif."""

import sys
from typing import List

import click


@click.command()
def main(args: List[str] = None) -> None:
    """Console script for adijif.

    Args:
        args (List[str]): List of input argument
    """
    click.echo("Replace this message by putting your code into " "adijif.cli.main")
    click.echo("See click documentation at https://click.palletsprojects.com/")


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""AD9088 Datapath Description Class."""


class ad9088_dp_rx:
    """AD9088 RX Data Path Configuration."""

    cddc_enabled = [True, True, True, True]
    cddc_decimations = [1, 1, 1, 1, 1, 1, 1, 1]
    cddc_decimations_available = [1, 2, 3, 4, 6, 12]
    cddc_nco_frequencies = [0, 0, 0, 0, 0, 0, 0, 0]
    cddc_nco_phases = [0, 0, 0, 0, 0, 0, 0, 0]

    fddc_enabled = [False] * 16
    fddc_decimations = [1] * 16
    fddc_decimations_available = [1, 2, 4, 8, 16, 32, 64]
    fddc_nco_frequencies = [0] * 16
    fddc_nco_phases = [0] * 16

    fddc_source = [1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8]

    def get_config(self) -> dict:
        """Get the datapath configuration for the AD9088 RX.

        Returns:
            dict: datapath configuration
        """
        datapath = {}
        datapath["cddc"] = {}
        datapath["cddc"]["enabled"] = self.cddc_enabled
        datapath["cddc"]["decimations"] = self.cddc_decimations
        datapath["cddc"]["nco_frequencies"] = self.cddc_nco_frequencies
        datapath["cddc"]["nco_phases"] = self.cddc_nco_phases

        datapath["fddc"] = {}
        datapath["fddc"]["enabled"] = self.fddc_enabled
        datapath["fddc"]["decimations"] = self.fddc_decimations
        datapath["fddc"]["nco_frequencies"] = self.fddc_nco_frequencies
        datapath["fddc"]["nco_phases"] = self.fddc_nco_phases
        datapath["fddc"]["source"] = self.fddc_source

        return datapath

    @property
    def decimation_overall(self) -> int:
        """Minimum Overall Decimation factor.

        Raises:
            Exception: No FDDC or CDDC enabled
            Exception: Enabled FDDC's source CDDC not enabled

        Returns:
            int: minimum overall decimation factor
        """
        if (not any(self.fddc_enabled)) and (not any(self.cddc_enabled)):
            raise Exception("No FDDCs or CDDCs enabled")

        if any(self.fddc_enabled):
            min_dec = -1
            for i, fdec in enumerate(self.fddc_decimations):
                if self.fddc_enabled[i]:
                    cddc = self.fddc_source[i]
                    if not self.cddc_enabled:
                        raise Exception(f"Source CDDC {cddc} not enabled for FDDC {i}")
                    cdec = self.cddc_decimations[cddc - 1]
                    if (cdec * fdec < min_dec) or min_dec == -1:
                        min_dec = cdec * fdec
        else:
            min_dec = -1
            for i, cdec in enumerate(self.cddc_decimations):
                if self.cddc_enabled[i] and (min_dec == -1 or cdec < min_dec):
                    min_dec = cdec

        return min_dec


# class ad9088_dp_tx:
#     """AD9088 TX Data Path Configuration."""

#     cduc_enabled = [True, True, True, True]
#     cduc_interpolation = 1
#     cduc_nco_frequencies = [0, 0, 0, 0]
#     cduc_nco_phases = [0, 0, 0, 0]

#     fduc_enabled = [False, False, False, False, False, False, False, False]
#     fduc_interpolation = 1
#     fduc_nco_frequencies = [0, 0, 0, 0, 0, 0, 0, 0]
#     fduc_nco_phases = [0, 0, 0, 0, 0, 0, 0, 0]

#     cduc_sources = [[1], [1], [3], [3]]

#     def get_config(self) -> dict:
#         """Get the datapath configuration for the AD9088 TX.

#         Returns:
#             dict: Datapath configuration
#         """
#         datapath = {}
#         datapath["cduc"] = {}
#         datapath["cduc"]["enabled"] = self.cduc_enabled
#         datapath["cduc"]["interpolation"] = self.cduc_interpolation
#         datapath["cduc"]["nco_frequencies"] = self.cduc_nco_frequencies
#         datapath["cduc"]["nco_phases"] = self.cduc_nco_phases
#         datapath["cduc"]["sources"] = self.cduc_sources

#         datapath["fduc"] = {}
#         datapath["fduc"]["enabled"] = self.fduc_enabled
#         datapath["fduc"]["interpolation"] = self.fduc_interpolation
#         datapath["fduc"]["nco_frequencies"] = self.fduc_nco_frequencies
#         datapath["fduc"]["nco_phases"] = self.fduc_nco_phases

#         return datapath



# flake8: noqa
import pytest

try:
    from gekko import GEKKO  # type: ignore
except ImportError:
    GEKKO = None


def skip_solver(solver):
    if solver.lower() == "gekko" and GEKKO is None:
        pytest.skip("Gekko not available")



# flake8: noqa
import os
import pprint

import pytest

import adijif
import adijif.converters
import adijif.converters.ad9084_util

here = os.path.dirname(os.path.abspath(__file__))
profile_json = os.path.join(
    here, "apollo_profiles", "ad9084_profiles", "id00_stock_mode.json"
)
summary_json = os.path.join(
    here, "apollo_profiles", "ad9084_profiles", "id00_stock_mode.summary"
)


ref = {
    "common_lane_rate_Hz": 20625000000,
    "core_clock_Hz": None,
    "datapath": {
        "cddc_decimation": 4,
        "cduc_interpolation": 4,
        "fddc_decimation": 2,
        "fduc_interpolation": 2,
    },
    "device_clock_Hz": 20000000000,
    "id": "id00_stock_mode",
    "is_8t8r": False,
    "jesd_settings": {
        "jrx": {
            "F": 1,
            "HD": True,
            # "K": 256,
            "L": 8,
            "M": 4,
            # "N": 16,
            "Np": 16,
            "S": 1,
        },
        "jtx": {
            "F": 1,
            "HD": True,
            # "K": 256,
            "L": 8,
            "M": 4,
            # "N": 16,
            "Np": 16,
            "S": 1,
        },
    },
    "profile_name": "id00_stock_mode",
    "rx_jesd_mode": 47,
    "tx_jesd_mode": 47,
}


def test_config_parser():

    settings = adijif.converters.ad9084_util.parse_json_config(profile_json)

    print("Settings:")
    pprint.pprint(settings)

    assert settings is not None

    # Verify the settings match the reference
    for key in ref:
        if isinstance(ref[key], dict):
            assert settings[key] == ref[key]
        else:
            assert settings[key] == ref[key], f"Mismatch for key: {key}"


def test_ad9084_separate_model_update():

    settings = adijif.converters.ad9084_util.parse_json_config(profile_json)

    print("Settings:")
    pprint.pprint(settings)

    assert settings is not None

    # Apply the settings to the AD9084 RX
    vcxo = 125000000
    clock = "hmc7044"
    sys = adijif.system("ad9084_rx", clock, "xilinx", vcxo, solver="CPLEX")

    adijif.converters.ad9084_util.apply_settings(sys.converter, settings)

    # Verify the settings were applied correctly
    # converter_rate / (cddc_dec * fddc_dec)
    assert sys.converter.converter_clock == settings["device_clock_Hz"]
    assert sys.converter.sample_clock == settings["device_clock_Hz"] / (
        settings["datapath"]["cddc_decimation"]
        * settings["datapath"]["fddc_decimation"]
    )
    # Check the JESD settings
    keys = settings["jesd_settings"]["jrx"].keys()
    for key in keys:
        assert (
            getattr(sys.converter, key) == settings["jesd_settings"]["jtx"][key]
        ), f"Mismatch for key: {key}"


def test_ad9084_model_update():

    settings = adijif.converters.ad9084_util.parse_json_config(profile_json)

    print("Settings:")
    pprint.pprint(settings)

    assert settings is not None

    # Apply the settings to the AD9084 RX
    vcxo = 125000000
    clock = "hmc7044"
    sys = adijif.system("ad9084_rx", clock, "xilinx", vcxo, solver="CPLEX")

    sys.converter.apply_profile_settings(profile_json)

    # Verify the settings were applied correctly
    # converter_rate / (cddc_dec * fddc_dec)
    assert sys.converter.converter_clock == settings["device_clock_Hz"]
    assert sys.converter.sample_clock == settings["device_clock_Hz"] / (
        settings["datapath"]["cddc_decimation"]
        * settings["datapath"]["fddc_decimation"]
    )
    # Check the JESD settings
    keys = settings["jesd_settings"]["jrx"].keys()
    for key in keys:
        assert (
            getattr(sys.converter, key) == settings["jesd_settings"]["jtx"][key]
        ), f"Mismatch for key: {key}"


def test_ad9088_smoke():

    vcxo = int(125e6)
    cddc_dec = 4
    fddc_dec = 4
    converter_rate = int(8e9)

    sys = adijif.system("ad9088_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    sys.fpga.setup_by_dev_kit_name("vcu118")
    sys.converter.sample_clock = converter_rate / (cddc_dec * fddc_dec)
    sys.converter.datapath.cddc_decimations = [cddc_dec] * 8
    sys.converter.datapath.fddc_decimations = [fddc_dec] * 16
    sys.converter.datapath.fddc_enabled = [True] * 16

    sys.converter.clocking_option = "direct"
    sys.add_pll_inline("adf4382", vcxo, sys.converter)
    sys.add_pll_sysref("adf4030", vcxo, sys.converter, sys.fpga)

    sys.clock.minimize_feedback_dividers = False

    mode_rx = adijif.utils.get_jesd_mode_from_params(
        sys.converter, M=8, L=8, S=1, Np=16, jesd_class="jesd204c"
    )

    sys.solve()


def test_ad9084_datapath_config():
    """Test AD9084 datapath configuration."""
    vcxo = 125000000
    clock = "hmc7044"
    sys = adijif.system("ad9084_rx", clock, "xilinx", vcxo, solver="CPLEX")

    # Configure datapath
    sys.converter.datapath.cddc_decimations = [4, 4, 4, 4]
    sys.converter.datapath.fddc_decimations = [2, 2, 2, 2, 2, 2, 2, 2]
    sys.converter.datapath.fddc_enabled = [True] * 8

    assert sys.converter.datapath.cddc_decimations == [4, 4, 4, 4]
    assert sys.converter.datapath.fddc_decimations == [2, 2, 2, 2, 2, 2, 2, 2]


def test_ad9088_datapath_config():
    """Test AD9088 datapath configuration."""
    vcxo = 125000000
    sys = adijif.system("ad9088_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    # Configure datapath
    sys.converter.datapath.cddc_decimations = [4] * 8
    sys.converter.datapath.fddc_decimations = [4] * 16
    sys.converter.datapath.fddc_enabled = [True] * 16

    assert len(sys.converter.datapath.cddc_decimations) == 8
    assert len(sys.converter.datapath.fddc_decimations) == 16


def test_ad9084_clocking_options():
    """Test AD9084 with different clocking options."""
    vcxo = 125000000
    sys = adijif.system("ad9084_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    # Test direct clocking (available for AD9084)
    sys.converter.clocking_option = "direct"
    assert sys.converter.clocking_option == "direct"


def test_ad9084_datapath_get_config():
    """Test AD9084 datapath get_config method."""
    vcxo = 125000000
    sys = adijif.system("ad9084_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    sys.converter.datapath.cddc_decimations = [4, 4, 4, 4]
    sys.converter.datapath.fddc_decimations = [2, 2, 2, 2, 2, 2, 2, 2]
    sys.converter.datapath.fddc_enabled = [True] * 8

    config = sys.converter.datapath.get_config()
    assert config is not None
    assert "cddc" in config
    assert "decimations" in config["cddc"]
    assert "fddc" in config
    assert "decimations" in config["fddc"]


def test_ad9088_datapath_get_config():
    """Test AD9088 datapath get_config method."""
    vcxo = 125000000
    sys = adijif.system("ad9088_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    sys.converter.datapath.cddc_decimations = [4] * 8
    sys.converter.datapath.fddc_decimations = [4] * 16
    sys.converter.datapath.fddc_enabled = [True] * 16

    config = sys.converter.datapath.get_config()
    assert config is not None


def test_ad9084_util_get_jesd_settings():
    """Test AD9084 util get_jesd_settings function."""
    import adijif.converters.ad9084_util as util

    # Test parsing settings
    settings = util.parse_json_config(profile_json)

    # Should have JESD settings
    assert "jesd_settings" in settings
    assert "jrx" in settings["jesd_settings"] or "jtx" in settings["jesd_settings"]


def test_ad9084_apply_settings_function():
    """Test AD9084 apply_settings utility function."""
    import adijif.converters.ad9084_util as util

    vcxo = 125000000
    sys = adijif.system("ad9084_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    settings = util.parse_json_config(profile_json)

    # Apply settings
    util.apply_settings(sys.converter, settings)

    # Verify settings were applied
    assert sys.converter.sample_clock > 0


def test_ad9084_converter_properties():
    """Test AD9084 converter properties."""
    vcxo = 125000000
    sys = adijif.system("ad9084_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    # Test various properties
    assert sys.converter.name == "AD9084_RX"
    assert hasattr(sys.converter, "datapath")
    assert hasattr(sys.converter, "clocking_option")


def test_ad9088_converter_properties():
    """Test AD9088 converter properties."""
    vcxo = 125000000
    sys = adijif.system("ad9088_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    # Test various properties
    assert sys.converter.name == "AD9088_RX"
    assert hasattr(sys.converter, "datapath")


def test_ad9084_sample_clock_setter():
    """Test AD9084 sample clock setter."""
    vcxo = 125000000
    sys = adijif.system("ad9084_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    # Set sample clock
    sys.converter.sample_clock = 2e9
    assert sys.converter.sample_clock == 2e9


def test_ad9088_sample_clock_setter():
    """Test AD9088 sample clock setter."""
    vcxo = 125000000
    sys = adijif.system("ad9088_rx", "hmc7044", "xilinx", vcxo, solver="CPLEX")

    # Set sample clock
    sys.converter.sample_clock = 2e9
    assert sys.converter.sample_clock == 2e9



# flake8: noqa
# pytype: skip-file
"""Common solver API management layer."""

from importlib.util import find_spec
from typing import Union

if find_spec("docplex"):
    from docplex.cp.expression import CpoExpr  # type: ignore
    from docplex.cp.expression import CpoFunctionCall  # type: ignore
    from docplex.cp.expression import CpoIntVar  # type: ignore
    from docplex.cp.model import binary_var  # type: ignore
    from docplex.cp.model import CpoModel, integer_var, interval_var  # type: ignore
    from docplex.cp.solution import CpoSolveResult  # type: ignore

    cplex_solver = True
else:
    cplex_solver = False
    CpoExpr = None
    CpoFunctionCall = None
    binary_var = None
    integer_var = None
    continuous_var = None
    interval_var = None

if find_spec("gekko"):
    import gekko  # type: ignore
    from gekko import GEKKO  # type: ignore
    from gekko.gk_operators import GK_Intermediate  # type: ignore
    from gekko.gk_operators import GK_Operators  # type: ignore
    from gekko.gk_variable import GKVariable  # type: ignore

    gekko_solver = True
else:
    gekko_solver = False
    gekko = None
    GEKKO = None
    GK_Intermediate = None
    GK_Operators = None
    GKVariable = None

if not cplex_solver and not gekko_solver:
    raise ImportError(
        "No solver found. gekko or docplex/cplex must be installed."
        + "\n-> Use `pip install pyadi-jif[cplex]` or `pip install pyadi-jif[gekko]`"
    )


def tround(value: float, tol: float = 1e-4) -> Union[float, int]:
    """Round if expected to have computational noise."""
    if value.is_integer():
        return int(value)
    if abs(value - round(value)) < tol:
        return round(value)
    return value



"""Ultrascale+ PLLs transceiver models."""

from typing import List, Union

from docplex.cp.modeler import if_then

from ...common import core
from ...converters.converter import converter as conv
from ...gekko_trans import gekko_translation
from ...solvers import CpoIntVar, GK_Intermediate, GK_Operators, GKVariable
from .pll import XilinxPLL
from .sevenseries import CPLL as SevenSeriesCPLL
from .sevenseries import QPLL as SevenSeriesQPLL


class UltraScalePlus(XilinxPLL, core, gekko_translation):
    """Ultrascale+ PLLs transceiver models."""

    # References
    # GTYs
    # https://docs.amd.com/v/u/en-US/ug578-ultrascale-gty-transceivers
    # https://docs.amd.com/r/en-US/ds925-zynq-ultrascale-plus/GTY-Transceiver-Switching-Characteristics
    # GTHs
    # https://docs.amd.com/v/u/en-US/ug576-ultrascale-gth-transceivers

    transceiver_types_available = ["GTHE4", "GTYE3", "GTYE4"]
    _transceiver_type = "GTHE4"

    force_cpll = False
    force_qpll = False
    force_qpll1 = False

    def add_plls(self) -> None:
        """Add PLLs to the model."""
        self.plls = {
            "CPLL": CPLL(self),
            "QPLL": QPLL(self),
            "QPLL1": QPLL1(self),
        }

    def add_constraints(
        self, config: dict, fpga_ref: Union[int, float], converter: conv
    ) -> dict:
        """Add constraints for PLLs.

        Args:
            config (dict): Configuration dictionary.
            fpga_ref (int, float): FPGA reference clock.
            converter (conv): Converter object.

        Returns:
            dict: Updated configuration dictionary.
        """
        assert self.plls, "No PLLs configured. Run the add_plls method"
        assert (
            self.force_cpll + self.force_qpll + self.force_qpll1 <= 1
        ), "Only one PLL can be enabled"
        for pll in self.plls:
            config = self.plls[pll].add_constraints(config, fpga_ref, converter)
        self._add_equation(
            config[converter.name + "_use_cpll"]
            + config[converter.name + "_use_qpll"]
            + config[converter.name + "_use_qpll1"]
            == 1
        )
        return config

    def get_config(
        self, config: dict, converter: conv, fpga_ref: Union[int, float]
    ) -> dict:
        """Get the configuration of the PLLs.

        Args:
            config (dict): Configuration dictionary.
            converter (conv): Converter object.
            fpga_ref (int, float): FPGA reference clock.

        Returns:
            dict: Updated configuration dictionary.
        """
        if self.force_cpll or self.force_qpll or self.force_qpll1:
            if self.force_cpll:
                ecpll = 1
                eqpll = self.solution.get_kpis()[converter.name + "_use_qpll"]
                eqpll1 = self.solution.get_kpis()[converter.name + "_use_qpll1"]
            elif self.force_qpll:
                ecpll = self.solution.get_kpis()[converter.name + "_use_cpll"]
                eqpll = 1
                eqpll1 = self.solution.get_kpis()[converter.name + "_use_qpll1"]
            else:
                ecpll = self.solution.get_kpis()[converter.name + "_use_cpll"]
                eqpll = self.solution.get_kpis()[converter.name + "_use_qpll"]
                eqpll1 = 1
        else:
            ecpll = self.solution.get_kpis()[converter.name + "_use_cpll"]
            eqpll = self.solution.get_kpis()[converter.name + "_use_qpll"]
            eqpll1 = self.solution.get_kpis()[converter.name + "_use_qpll1"]
        assert ecpll + eqpll + eqpll1 == 1, "Only one PLL can be enabled"
        if ecpll:
            pll = "CPLL"
        elif eqpll:
            pll = "QPLL"
        else:
            pll = "QPLL1"
        return self.plls[pll].get_config(config, converter, fpga_ref)


class CPLL(SevenSeriesCPLL):
    """CPLL model for Ultrascale+ transceivers."""

    @property
    def vco_min(self) -> int:
        """Get the VCO min frequency in Hz for CPLL."""
        if self.parent.transceiver_type in ["GTHE4", "GTYE3", "GTYE4"]:
            return 2000000000
        raise Exception(
            f"Unknown vco_min for transceiver type {self.parent.transceiver_type}"
        )

    @property
    def vco_max(self) -> int:
        """Get the VCO max frequency in Hz for CPLL."""
        if self.parent.transceiver_type in ["GTHE4", "GTYE3", "GTYE4"]:
            return 6250000000
        raise Exception(
            f"Unknown vco_max for transceiver type {self.parent.transceiver_type}"
        )


class QPLL(SevenSeriesQPLL):
    """QPLL model for Ultrascale+ transceivers."""

    force_integer_mode = True

    @property
    def vco_min(self) -> int:
        """Get the VCO min frequency in Hz for QPLL."""
        if self.parent.transceiver_type in ["GTHE3", "GTHE4", "GTYE4"]:
            return 9800000000
        raise Exception(
            f"Unknown vco_min for transceiver type {self.parent.transceiver_type}"
        )

    @property
    def vco_max(self) -> int:
        """Get the VCO max frequency in Hz for QPLL."""
        if self.parent.transceiver_type in ["GTHE3", "GTHE4", "GTYE4"]:
            return 16375000000
        raise Exception(
            f"Unknown vco_max for transceiver type {self.parent.transceiver_type}"
        )

    N_available = [*range(16, 160 + 1)]
    _N = [*range(16, 160 + 1)]

    D_available = [1, 2, 4, 8, 16]
    _D = [1, 2, 4, 8, 16]
    # 32 not available in AC modes https://docs.amd.com/r/en-US/ds925-zynq-ultrascale-plus/GTY-Transceiver-Switching-Characteristics # type: ignore # noqa: B950

    @property
    def QPLL_CLKOUTRATE_available(self) -> List[int]:
        """Get the QPLL_CLKOUTRATE available values."""
        if self.parent.transceiver_type == "GTHE4":
            return [1, 2]
        return [1]

    _QPLL_CLKOUTRATE_GTY = [1, 2]
    _QPLL_CLKOUTRATE_GTH = [1, 2]

    @property
    def QPLL_CLKOUTRATE(self) -> int:
        """Get the QPLL_CLKOUTRATE value."""
        if "GTH" in self.parent.transceiver_type:
            return self._QPLL_CLKOUTRATE_GTH
        return self._QPLL_CLKOUTRATE_GTY

    @QPLL_CLKOUTRATE.setter
    def QPLL_CLKOUTRATE(self, val: int) -> None:
        """Set the QPLL_CLKOUTRATE.

        Args:
            val (int): QPLL_CLKOUTRATE value.

        Raises:
            ValueError: If QPLL_CLKOUTRATE is out of range.
        """
        self._check_in_range(val, self.QPLL_CLKOUTRATE_available, "QPLL_CLKOUTRATE")
        if "GTH" in self.parent.transceiver_type:
            self._QPLL_CLKOUTRATE_GTH = val
        raise ValueError(
            f"QPLL_CLKOUTRATE not available for {self.parent.transceiver_type}"
        )

    SDMDATA_min_max = [0, 2**24 - 1]
    _SDMDATA_min = 0

    @property
    def SDMDATA_min(self) -> int:
        """Get the SDMDATA_min value."""
        return self._SDMDATA_min

    @SDMDATA_min.setter
    def SDMDATA_min(self, val: int) -> None:
        """Set the SDMDATA_min.

        Args:
            val (int): SDMDATA_min value.

        Raises:
            ValueError: If SDMDATA_min is out of range.
        """
        if val < self.SDMDATA_min_max[0] or val > self.SDMDATA_min_max[1]:
            raise ValueError(
                f"SDMDATA_min must be between {self.SDMDATA_min_max[0]} and"
                + f" {self.SDMDATA_min_max[1]}"
            )
        self._SDMDATA_min = val

    _SDMDATA_max = 2**24 - 1

    @property
    def SDMDATA_max(self) -> int:
        """Get the SDMDATA_max value."""
        return self._SDMDATA_max

    @SDMDATA_max.setter
    def SDMDATA_max(self, val: int) -> None:
        """Set the SDMDATA_max.

        Args:
            val (int): SDMDATA_max value.

        Raises:
            ValueError: If SDMDATA_max is out of range.
        """
        if val < self.SDMDATA_min_max[0] or val > self.SDMDATA_min_max[1]:
            raise ValueError(
                f"SDMDATA must be between {self.SDMDATA_min_max[0]} and"
                + f" {self.SDMDATA_min_max[1]}"
            )
        self._SDMDATA_max = val

    SDMWIDTH_available = [16, 20, 24]
    _SDMWIDTH = [16, 20, 24]

    @property
    def SDMWIDTH(self) -> int:
        """Get the SDMWIDTH value."""
        return self._SDMWIDTH

    @SDMWIDTH.setter
    def SDMWIDTH(self, val: int) -> None:
        """Set the SDMWIDTH.

        Args:
            val (int): SDMWIDTH value.
        """
        self._check_in_range(val, self.SDMWIDTH_available, "SDMWIDTH")
        self._SDMWIDTH = val

    _pname = "qpll"

    def get_config(
        self, config: dict, converter: conv, fpga_ref: Union[int, float]
    ) -> dict:
        """Get the configuration of the QPLL.

        Args:
            config (dict): Configuration dictionary.
            converter (conv): Converter object.
            fpga_ref (int, float): FPGA reference clock.

        Returns:
            dict: Updated configuration dictionary.
        """
        pname = self._pname
        pll_config = {"type": self._pname}
        pll_config["n"] = self._get_val(config[converter.name + f"_n_{pname}"])
        pll_config["m"] = self._get_val(config[converter.name + f"_m_{pname}"])
        pll_config["d"] = self._get_val(config[converter.name + f"_d_{pname}"])
        pll_config["clkout_rate"] = self._get_val(
            config[converter.name + f"_clkout_rate_{pname}"]
        )
        if converter.bit_clock < 28.1e9 and not self.force_integer_mode:
            sdm_data = self._get_val(config[converter.name + f"_sdm_data_{pname}"])
            if sdm_data > 0:
                pll_config["sdm_data"] = sdm_data
                pll_config["sdm_width"] = self._get_val(
                    config[converter.name + f"_sdm_width_{pname}"]
                )
                pll_config["frac"] = self.solution.get_kpis()[
                    converter.name + f"_frac_{pname}"
                ]
                pll_config["n_dot_frac"] = self.solution.get_kpis()[
                    converter.name + f"_n_dot_frac_{pname}"
                ]
            else:
                pll_config["n_dot_frac"] = pll_config["n"]
        else:
            pll_config["n_dot_frac"] = pll_config["n"]

        pll_config["n"] = pll_config["n_dot_frac"]

        # config['vco'] = self._get_val(config[converter.name + f"_vco_{pname}"])
        pll_config["vco"] = self.solution.get_kpis()[converter.name + f"_vco_{pname}"]

        # Check
        pll_out = (
            fpga_ref
            * pll_config["n_dot_frac"]
            / (pll_config["m"] * pll_config["clkout_rate"])
        )
        lane_rate = pll_out * 2 / pll_config["d"]
        assert lane_rate == converter.bit_clock, f"{lane_rate} != {converter.bit_clock}"

        return pll_config

    def add_constraints(
        self,
        config: dict,
        fpga_ref: Union[int, GKVariable, GK_Intermediate, GK_Operators, CpoIntVar],
        converter: conv,
    ) -> dict:
        """Add constraints for the Transceiver.

        Args:
            config (dict): Configuration dictionary.
            fpga_ref (int, CpoIntVar): FPGA reference clock.
            converter (conv): Converter object.

        Returns:
            dict: Updated configuration dictionary.
        """
        pname = self._pname

        # Global flag to use QPLLn
        if self.parent.force_qpll and self._pname == "qpll":
            v = 1
        elif self.parent.force_qpll1 and self._pname == "qpll1":
            v = 1
        elif self.parent.force_cpll and self._pname == "cpll":
            v = 1
        else:
            v = [0, 1]

        config[converter.name + f"_use_{pname}"] = self._convert_input(
            v, converter.name + f"_use_{pname}"
        )
        if v == [0, 1]:
            self.model.add_kpi(
                config[converter.name + f"_use_{pname}"],
                name=converter.name + f"_use_{pname}",
            )

        # Add variables
        config[converter.name + f"_m_{pname}"] = self._convert_input(
            self.M, converter.name + f"_m_{pname}"
        )
        config[converter.name + f"_d_{pname}"] = self._convert_input(
            self.D, converter.name + f"_d_{pname}"
        )
        config[converter.name + f"_n_{pname}"] = self._convert_input(
            self.N, converter.name + f"_n_{pname}"
        )
        config[converter.name + f"_clkout_rate_{pname}"] = self._convert_input(
            self.QPLL_CLKOUTRATE, converter.name + f"_clkout_rate_{pname}"
        )

        # Frac part
        if converter.bit_clock < 28.1e9 and not self.force_integer_mode:
            config[converter.name + f"_sdm_data_{pname}"] = self.model.integer_var(
                min=self.SDMDATA_min,
                max=self.SDMDATA_max,
                name=converter.name + f"_sdm_data_{pname}",
            )
            config[converter.name + f"_sdm_width_{pname}"] = self._convert_input(
                self.SDMWIDTH, converter.name + f"_sdm_width_{pname}"
            )
            config[converter.name + f"_HIGH_RATE_{pname}"] = self._convert_input(
                self.QPLL_CLKOUTRATE, converter.name + f"_HIGH_RATE_{pname}"
            )

        # Add intermediate variables
        if converter.bit_clock < 28.1e9 and not self.force_integer_mode:
            config[converter.name + f"_frac_{pname}"] = self._add_intermediate(
                config[converter.name + f"_sdm_data_{pname}"]
                / (2 ** config[converter.name + f"_sdm_width_{pname}"])
            )
            self.model.add_kpi(
                config[converter.name + f"_frac_{pname}"],
                name=converter.name + f"_frac_{pname}",
            )
            self._add_equation([config[converter.name + f"_frac_{pname}"] < 1])
            config[converter.name + f"_n_dot_frac_{pname}"] = self._add_intermediate(
                config[converter.name + f"_n_{pname}"]
                + config[converter.name + f"_frac_{pname}"]
            )
            self.model.add_kpi(
                config[converter.name + f"_n_dot_frac_{pname}"],
                name=converter.name + f"_n_dot_frac_{pname}",
            )
        else:
            config[converter.name + f"_n_dot_frac_{pname}"] = self._add_intermediate(
                config[converter.name + f"_n_{pname}"]
            )

        config[converter.name + f"_pll_out_{pname}"] = self._add_intermediate(
            fpga_ref
            * config[converter.name + f"_n_dot_frac_{pname}"]
            / (
                config[converter.name + f"_m_{pname}"]
                * config[converter.name + f"_clkout_rate_{pname}"]
            )
        )
        config[converter.name + f"_vco_{pname}"] = self._add_intermediate(
            fpga_ref
            * config[converter.name + f"_n_dot_frac_{pname}"]
            / (
                config[converter.name + f"_m_{pname}"]
                * config[converter.name + f"_clkout_rate_{pname}"]
            )
        )
        self.model.add_kpi(
            config[converter.name + f"_vco_{pname}"],
            name=converter.name + f"_vco_{pname}",
        )

        # Add constraints
        self._add_equation(
            [
                if_then(
                    config[converter.name + f"_use_{pname}"] == 1,
                    converter.bit_clock * config[converter.name + f"_d_{pname}"]
                    == config[converter.name + f"_pll_out_{pname}"] * 2,
                ),
                # if_then(
                #     config[converter.name + f"_use_{pname}"] == 1,
                #     config[converter.name + f"_HIGH_RATE_{pname}"]
                #     == (converter.bit_clock >= 28.1e9),
                # ),
            ]
        )

        self._add_equation(
            [
                if_then(
                    config[converter.name + f"_use_{pname}"] == 1,
                    config[converter.name + f"_vco_{pname}"] >= self.vco_min,
                ),
                if_then(
                    config[converter.name + f"_use_{pname}"] == 1,
                    config[converter.name + f"_vco_{pname}"] <= self.vco_max,
                ),
                # if_then(
                #     config[converter.name + f"_HIGH_RATE_{pname}"] == 1,
                #     config[converter.name + f"_frac_{pname}"] == 0,
                # ),
            ]
        )

        return config


class QPLL1(QPLL):
    """QPLL1 model for Ultrascale+ transceivers."""

    _pname = "qpll1"

    @property
    def vco_min(self) -> int:
        """VCO min frequency in Hz for QPLL1."""
        if self.parent.transceiver_type in ["GTHE3", "GTHE4", "GTYE3", "GTYE4"]:
            return 8000000000
        raise Exception(
            f"Unknown vco_min for transceiver type {self.parent.transceiver_type}"
        )

    @property
    def vco_max(self) -> int:
        """VCO max frequency in Hz for QPLL1."""
        if self.parent.transceiver_type in ["GTHE3", "GTHE4", "GTYE3", "GTYE4"]:
            return 13000000000
        raise Exception(
            f"Unknown vco_max for transceiver type {self.parent.transceiver_type}"
        )



"""ADI JIF utility types and functions."""

from typing import Dict, Union

import numpy as np

from adijif.solvers import GEKKO, CpoModel, integer_var  # type: ignore


class range:
    """Model range of possible values.

    This class is similar to range but interfaces with
    the modeling layers to create correct mappings for internal
    solvers
    """

    def __init__(self, start: int, stop: int, step: int, name: str) -> None:
        """Model range of possible values.

        This class is similar to range but interfaces with
        the modeling layers to create correct mappings for internal
        solvers

        Args:
            start (int): Start value of range
            stop (int): Last value of range-step
            step (int): Difference between successive values in range
            name (str): Name of variable
        """
        assert isinstance(start, int), "start must be an int"
        assert isinstance(stop, int), "stop must be an int"
        assert isinstance(step, int), "step must be an int"
        assert isinstance(name, str), "name must be a string"

        self.start = start
        self.stop = stop
        self.step = step
        self.name = name

    def __call__(self, model: Union[GEKKO, CpoModel]) -> Dict:
        """Generate range for parameter solver.

        Args:
            model (GEKKO, CpoModel): Model of JESD system or part to solve

        Returns:
            Dict: Dictionary of solver variable(s) for model
        """
        if GEKKO and CpoModel:
            assert isinstance(
                model, (GEKKO, CpoModel)
            ), "range must be called with input type model"
        elif GEKKO:
            assert isinstance(
                model, GEKKO
            ), "range must be called with input type model"
        elif CpoModel:
            assert isinstance(
                model, CpoModel
            ), "range must be called with input type model"

        config = {}
        if isinstance(model, CpoModel):
            o_array = np.arange(self.start, self.stop, self.step)
            config["range"] = integer_var(domain=o_array, name=self.name + "_Var")
            return config

        if self.step == 1:
            config["range"] = model.Var(
                integer=True,
                lb=self.start,
                ub=self.stop,
                value=self.start,
                name=self.name + "_Var",
            )
        else:
            # Create piecewise problem with binary values (DO NOT USE SOS1!)
            o_array = np.arange(self.start, self.stop, self.step)
            o_array = list(map(int, o_array))

            options = model.Array(model.Var, len(o_array), lb=0, ub=1, integer=True)
            model.Equation(model.sum(options) == 1)  # only select one
            config["range"] = model.Intermediate(model.sum(o_array * options))

        return config


class arb_source:
    """Arbitrary source for solver.

    This internally uses a division of two integers to create an arbitrary clock
    source. This allows the solver to find optimal rational values for reference
    clocks that can be any frequency within a continuous range.

    IMPORTANT: Currently only supports CPLEX solver (CpoModel). GEKKO solver is
    not supported. Use the 'range' type for discrete values with GEKKO.

    Example:
        # Create arbitrary VCXO between 100-200 MHz
        vcxo = adijif.types.arb_source("vcxo", a_min=100000000, a_max=200000000,
                                       b_min=1, b_max=1)

        # Or let solver find optimal rational frequency
        vcxo = adijif.types.arb_source("vcxo", a_min=1, a_max=int(1e11),
                                       b_min=1, b_max=int(1e11))

        # Use with system (requires CPLEX)
        sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    """

    _max_scalar = int(1e11)

    def __init__(
        self,
        name: str,
        a_min: Union[None, int] = None,
        b_min: Union[None, int] = None,
        a_max: Union[None, int] = None,
        b_max: Union[None, int] = None,
    ) -> None:
        """Arbitrary source for solver.

        Args:
            name (str): Name of source
            a_min (int, optional): Minimum value for numerator. Defaults to None.
            b_min (int, optional): Minimum value for denominator. Defaults to None.
            a_max (int, optional): Maximum value for numerator. Defaults to None.
            b_max (int, optional): Maximum value for denominator. Defaults to None.
        """
        self.name = name
        if a_min is None:
            a_min = 0
        if a_max is None:
            a_max = self._max_scalar
        if b_min is None:
            b_min = 0
        if b_max is None:
            b_max = self._max_scalar

        self._a = integer_var(a_min, a_max, name=name + "_a")
        self._b = integer_var(b_min, b_max, name=name + "_b")

    def __call__(self, model: Union[GEKKO, CpoModel]) -> Dict:
        """Generate arbitrary source for solver.

        Args:
            model (GEKKO, CpoModel): Model of JESD system or part to solve

        Returns:
            Dict: Dictionary of solver variable(s) for model

        Raises:
            NotImplementedError: Only CpoModel (CPLEX) is supported
        """
        if GEKKO and CpoModel:
            assert isinstance(
                model, (GEKKO, CpoModel)
            ), "arb_source must be called with input type model"
        elif GEKKO:
            assert isinstance(
                model, GEKKO
            ), "arb_source must be called with input type model"
        elif CpoModel:
            assert isinstance(
                model, CpoModel
            ), "arb_source must be called with input type model"

        # config = {}
        if isinstance(model, CpoModel):
            # config[self.name] = self._a / self._b
            # return config
            return self._a / self._b

        raise NotImplementedError(
            "arb_source only supports CPLEX solver (CpoModel). "
            "For GEKKO solver, use adijif.types.range for discrete values instead. "
            "Install CPLEX with: pip install pyadi-jif[cplex]"
        )

    def get_config(self, solution: Dict) -> Dict:
        """Get configuration from solver results.

        Args:
            solution (Dict): Solver results

        Returns:
            Dict: Dictionary of solver variable(s) for model
        """
        return {self.name: solution[self._a] / solution[self._b]}



# flake8: noqa

import pytest

import adijif


def test_converter_quick_config_with_invalid_mode():
    """Test quick configuration with invalid mode."""
    conv = adijif.ad9680()

    with pytest.raises(
        Exception, match="Mode .* not among configurations for jesd204.*"
    ):
        conv.set_quick_configuration_mode("INVALID_MODE", "jesd204b")


def test_converter_quick_config_with_invalid_jesd_class():
    """Test quick configuration with invalid JESD class."""
    conv = adijif.ad9680()

    with pytest.raises(Exception, match=".* not available for JESD class for .*"):
        conv.set_quick_configuration_mode("0x88", "invalid_jesd_class")


def test_converter_apply_quick_config_with_mode_number():
    """Test applying quick config with mode number instead of string."""
    conv = adijif.ad9680()

    # Test with integer mode
    conv.set_quick_configuration_mode(0x88, "jesd204b")
    assert conv.jesd_class == "jesd204b"


def test_converter_device_clock_available():
    """Test device_clock_available method."""
    conv = adijif.ad9680()

    conv.sample_clock = 1e9
    conv.decimation = 1

    clocks = conv.device_clock_available()
    assert clocks is not None
    assert len(clocks) > 0


def test_converter_get_config():
    """Test get_config method returns proper structure."""
    conv = adijif.ad9680()

    conv.sample_clock = 1e9
    conv.decimation = 1
    conv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Create solution
    required_clocks = conv.get_required_clocks()
    required_clock_names = conv.get_required_clock_names()

    # Add generic clock sources
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):
        clk = adijif.types.arb_source(name)
        conv._add_equation(clk(conv.model) == clock)
        clks.append(clk)

    solution = conv.model.solve(LogVerbosity="Quiet")
    config = conv.get_config(solution)

    assert config is not None
    # assert "jesd_class" in config
    # assert "sample_clock" in config


def test_converter_get_jesd_config():
    """Test get_jesd_config method."""
    conv = adijif.ad9680()

    conv.sample_clock = 1e9
    conv.decimation = 1
    conv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Create solution
    required_clocks = conv.get_required_clocks()
    required_clock_names = conv.get_required_clock_names()

    # Add generic clock sources
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):
        clk = adijif.types.arb_source(name)
        conv._add_equation(clk(conv.model) == clock)
        clks.append(clk)

    solution = conv.model.solve(LogVerbosity="Quiet")
    jesd_config = conv.get_jesd_config(solution)

    assert jesd_config is not None
    assert "L" in jesd_config
    assert "M" in jesd_config


def test_converter_clocking_option_invalid():
    """Test setting invalid clocking option."""
    conv = adijif.ad9680()

    with pytest.raises(Exception, match="clocking_option not available"):
        conv.clocking_option = "invalid_option"


def test_converter_clocking_option_valid():
    """Test setting valid clocking option."""
    conv = adijif.ad9680()

    conv.clocking_option = "direct"
    assert conv.clocking_option == "direct"


def test_converter_name_property():
    """Test converter name property."""
    conv = adijif.ad9680()

    # Default name
    assert conv.name == "AD9680"

    # Set custom name
    conv.name = "custom_name"
    assert conv.name == "custom_name"


def test_converter_multiframe_clock():
    """Test multiframe_clock property."""
    conv = adijif.ad9680()

    conv.sample_clock = 1e9
    conv.decimation = 1
    conv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Multiframe clock should be calculated
    mfc = conv.multiframe_clock
    assert mfc is not None
    assert mfc > 0


def test_converter_bit_clock():
    """Test bit_clock property."""
    conv = adijif.ad9680()

    conv.sample_clock = 1e9
    conv.decimation = 1
    conv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Bit clock should be calculated
    bc = conv.bit_clock
    assert bc is not None
    assert bc > 0


def test_converter_validate_config():
    """Test validate_config method."""
    conv = adijif.ad9680()

    conv.sample_clock = 1e9
    conv.decimation = 1
    conv.set_quick_configuration_mode(str(0x88), "jesd204b")

    # Should not raise exception
    conv.validate_config()


def test_converter_validate_config_invalid():
    """Test validate_config with invalid parameters."""
    conv = adijif.ad9680()

    conv.sample_clock = 1e9
    conv.decimation = 1
    # Don't set JESD mode - should have invalid defaults

    with pytest.raises(Exception, match="K not in range for device"):
        # Set invalid K value
        conv.K = 2000
        # conv.validate_config()


def test_converter_sample_clock_property():
    """Test sample_clock property."""
    conv = adijif.ad9680()

    # Set sample clock
    conv.sample_clock = 1e9
    assert conv.sample_clock == 1e9


def test_converter_available_jesd_modes():
    """Test available_jesd_modes property."""
    conv = adijif.ad9680()

    modes = conv.available_jesd_modes
    assert modes is not None
    assert len(modes) > 0


def test_converter_nested_property():
    """Test _nested property for nested converters."""
    # Test non-nested converter
    conv = adijif.ad9680()
    assert conv._nested == False

    # Test nested converter
    conv_nested = adijif.ad9081()
    assert conv_nested._nested is not None
    assert len(conv_nested._nested) > 0



"""Page definitions for PyADI-JIF Tools Explorer."""

from typing import Dict, Type

from ..utils import Page
from .clockconfigurator import ClockConfigurator
from .jesdmodeselector import JESDModeSelector
from .systemconfigurator import SystemConfigurator

PAGE_MAP: Dict[str, Type[Page]] = {
    "JESD204 Mode Selector": JESDModeSelector,
    "Clock Configurator": ClockConfigurator,
    "System Configurator": SystemConfigurator,
}

__all__ = ["PAGE_MAP"]



"""Drawing features for Xilinx FPGA designs."""

from typing import Dict, List

from adijif.converters.converter import converter as Converter
from adijif.draw import Layout, Node


class xilinx_draw:
    """Xilinx drawing features."""

    def _init_diagram(self) -> None:
        """Initialize the diagram for a Xilinx FPGA alone."""
        self.ic_diagram_node = None

        self.ic_diagram_node = Node(self.name)

        # Add generic transceiver since we don't know what is used until later
        transceiver = Node("Transceiver", ntype="transceiver")
        self.ic_diagram_node.add_child(transceiver)

        # Add Link layer JESD204 core
        jesd204_link = Node("JESD204-Link-IP", ntype="ip")
        self.ic_diagram_node.add_child(jesd204_link)

        # Add Transport Layer JESD204 core
        jesd204_transport = Node("JESD204-Transport-IP", ntype="ip")
        self.ic_diagram_node.add_child(jesd204_transport)

        # Add Application layer JESD204 core
        jesd204_application = Node("JESD204-Application-IP", ntype="ip")
        self.ic_diagram_node.add_child(jesd204_application)

        # Add connections
        self.ic_diagram_node.add_connection({"from": transceiver, "to": jesd204_link})
        self.ic_diagram_node.add_connection(
            {"from": jesd204_link, "to": jesd204_transport}
        )
        self.ic_diagram_node.add_connection(
            {"from": jesd204_transport, "to": jesd204_application, "type": "data"}
        )

    def _draw_phy(self, config: Dict, converter: Converter = None) -> tuple:
        """Draw the PHY layer for the Xilinx FPGA.

        Args:
            config (Dict): Configuration dictionary
            converter (Converter): Converter object (optional)

        Returns:
            tuple: Tuple of in_c, out_c, connect_to_input

        Raises:
            Exception: If unknown out_clk_select
        """
        # cfg = {
        #     "clocks": {"FPGA_REF": 500000000.0, "LINK_OUT_REF": 125000000.0},
        #     "fpga": {
        #         "type": "cpll",
        #         "m": 2,
        #         "d": 1,
        #         "n1": 5,
        #         "n2": 4,
        #         "vco": 5000000000.0,
        #         "sys_clk_select": "XCVR_CPLL",
        #         "progdiv": 40.0,
        #         "out_clk_select": "XCVR_PROGDIV_CLK",
        #         "separate_device_clock_required": 1,
        #         "transport_samples_per_clock": 8,
        #     },
        # }

        if converter:
            name = f"{self.name}_{converter.name.upper()}_ref_clk"
            ref_in_rate = config["clocks"][name]
        else:
            ref_in_rate = config["clocks"]["FPGA_REF"]

        phy = Node("JESD204-PHY-IP", ntype="phy")
        self.ic_diagram_node.add_child(phy)

        # PLL
        if config["fpga"]["type"] == "cpll":
            cpll = Node("CPLL", ntype="cpll")
            phy.add_child(cpll)

            # Put stuff in CPLL
            m = Node("M", ntype="divider")
            cpll.add_child(m)
            m.value = config["fpga"]["m"]

            pfd = Node("PFD", ntype="phase-frequency-detector")
            cpll.add_child(pfd)
            cpll.add_connection({"from": m, "to": pfd, "rate": ref_in_rate})

            cp = Node("CP", ntype="charge-pump")
            cpll.add_child(cp)
            cpll.add_connection({"from": pfd, "to": cp})

            lpf = Node("LPF", ntype="loop-filter")
            cpll.add_child(lpf)
            cpll.add_connection({"from": cp, "to": lpf})

            vco = Node("VCO", ntype="vco")
            cpll.add_child(vco)
            cpll.add_connection({"from": lpf, "to": vco})

            d = Node("D", ntype="divider")
            cpll.add_child(d)
            d.value = config["fpga"]["d"]
            cpll.add_connection({"from": vco, "to": d, "rate": config["fpga"]["vco"]})

            n1 = Node("N1", ntype="divider")
            cpll.add_child(n1)
            n1.value = config["fpga"]["n1"]
            cpll.add_connection({"from": vco, "to": n1})

            n2 = Node("N2", ntype="divider")
            cpll.add_child(n2)
            n2.value = config["fpga"]["n2"]
            cpll.add_connection({"from": n1, "to": n2})
            cpll.add_connection({"from": n2, "to": pfd})

            transceiver = Node("Transceiver", ntype="transceiver")
            phy.add_child(transceiver)
            phy.add_connection({"from": d, "to": transceiver})

            # Define common in/out
            in_c = m
            xcvr_out = d
            xcvr_out_rate = config["fpga"]["vco"] / config["fpga"]["d"]

        else:

            qpll = Node("QPLL", ntype="qpll")
            phy.add_child(qpll)

            # Put stuff in QPLL
            m = Node("M", ntype="divider")
            m.value = config["fpga"]["m"]
            qpll.add_child(m)

            pfd = Node("PFD", ntype="phase-frequency-detector")
            qpll.add_child(pfd)

            cp = Node("CP", ntype="charge-pump")
            qpll.add_child(cp)

            lpf = Node("LPF", ntype="loop-filter")
            qpll.add_child(lpf)

            vco = Node("VCO", ntype="vco")
            qpll.add_child(vco)

            n = Node("N", ntype="divider")
            n.value = config["fpga"]["n"]
            qpll.add_child(n)

            # d2 = Node("/2", ntype="divider")
            # qpll.add_child(d2)

            d = Node("D", ntype="divider")
            d.value = config["fpga"]["d"]
            qpll.add_child(d)

            # Connect
            phy.add_connection({"from": m, "to": pfd})
            phy.add_connection({"from": pfd, "to": cp})
            phy.add_connection({"from": cp, "to": lpf})
            phy.add_connection({"from": lpf, "to": vco})
            phy.add_connection({"from": vco, "to": n})
            # TRX is DDR devices so it uses both clock edges
            #   (skipping /2 and *2 dividers)
            # phy.add_connection({"from": vco, "to": d2,
            #   "rate": config['fpga']['vco']})
            # phy.add_connection({"from": d2, "to": d,
            #   "rate": config['fpga']['vco'] / 2})
            phy.add_connection({"from": vco, "to": d, "rate": config["fpga"]["vco"]})
            phy.add_connection({"from": n, "to": pfd})

            xcvr_out = d
            xcvr_out_rate = config["fpga"]["vco"] / config["fpga"]["d"]

            in_c = m

        # Divider complex
        trx_dividers = Node("Transceiver Dividers", ntype="trx-dividers")
        phy.add_child(trx_dividers)

        if config["fpga"]["out_clk_select"] == "XCVR_OUTCLK_PCS":
            raise Exception("Only XCVR_PROGDIV_CLK supported for now")
        elif config["fpga"]["out_clk_select"] == "XCVR_OUTCLK_PMA":
            raise Exception("Only XCVR_PROGDIV_CLK supported for now")
        elif config["fpga"]["out_clk_select"] == "XCVR_REFCLK":

            rmux = Node("REFCLKSEL-Mux", ntype="mux")
            trx_dividers.add_child(rmux)
            connect_to_input = [rmux]

            smux = Node("SYSCLKSEL-Mux", ntype="mux")
            trx_dividers.add_child(smux)
            trx_dividers.add_connection({"from": rmux, "to": smux, "rate": ref_in_rate})

            out_rate = ref_in_rate
            out = smux

        elif config["fpga"]["out_clk_select"] == "XCVR_REFCLK_DIV2":

            rmux = Node("REFCLKSEL-Mux", ntype="mux")
            trx_dividers.add_child(rmux)
            # trx_dividers.add_connection(
            #     {"from": in_c, "to": rmux, "rate": ref_in_rate}
            # )
            connect_to_input = [rmux]

            smux = Node("SYSCLKSEL-Mux", ntype="mux")
            trx_dividers.add_child(smux)
            trx_dividers.add_connection({"from": rmux, "to": smux, "rate": ref_in_rate})

            div2 = Node("DIV2", ntype="divider")
            trx_dividers.add_child(div2)
            div2.value = 2
            trx_dividers.add_connection({"from": smux, "to": div2, "rate": ref_in_rate})

            out_rate = ref_in_rate / 2
            out = div2

        elif config["fpga"]["out_clk_select"] == "XCVR_PROGDIV_CLK":

            mux = Node("PLLCLKSEL-Mux", ntype="mux")
            mux.value = config["fpga"]["type"]
            trx_dividers.add_child(mux)
            phy.add_connection({"from": xcvr_out, "to": mux, "rate": xcvr_out_rate})

            cdr = Node("CDR", ntype="cdr")
            trx_dividers.add_child(cdr)
            trx_dividers.add_connection({"from": mux, "to": cdr})

            progdiv = Node("ProgDiv", ntype="divider")
            trx_dividers.add_child(progdiv)
            progdiv.value = int(config["fpga"]["progdiv"])
            trx_dividers.add_connection({"from": cdr, "to": progdiv})

            out_rate = xcvr_out_rate / config["fpga"]["progdiv"]
            out = progdiv
            connect_to_input = []
        else:
            raise Exception(
                f"Unknown out_clk_select: {config['fpga']['out_clk_select']}"
            )

        out_mux = Node("OUTCLKSEL-Mux", ntype="mux")
        trx_dividers.add_child(out_mux)
        trx_dividers.add_connection({"from": out, "to": out_mux, "rate": out_rate})

        return in_c, out_mux, connect_to_input, xcvr_out

    def draw(
        self, config: Dict, lo: Layout = None, converters: List[Converter] = None
    ) -> str:
        """Draw diagram in d2 language for IC alone with reference clock.

        Args:
            config (Dict): Clock settings
            lo (Layout): Layout object to draw on (optional)
            converters (List): List of converters to draw (optional)

        Returns:
            str: SVG data

        Raises:
            Exception: If no solution is saved
            Exception: If no converter is found
        """
        if not self._saved_solution:
            raise Exception("No solution to draw. Must call solve first.")

        system_draw = lo is not None

        if not system_draw:
            lo = Layout(f"{self.name} Example")
            converters = []
        else:
            # Verify lo is a Layout object
            assert isinstance(lo, Layout), "lo must be a Layout object"
            assert len(converters) == 1, "Only one converter supported"

        lo.add_node(self.ic_diagram_node)

        clocks = config["clocks"]
        device_clock_source = config["fpga"]["device_clock_source"]

        if not system_draw:
            ref_in = Node("REF_IN", ntype="input")
            lo.add_node(ref_in)
        else:
            for converter in converters:
                to_node = lo.get_node(f"{self.name}_{converter.name.upper()}_ref_clk")
                # Locate node connected to this one
                from_node = lo.get_connection(to=to_node.name)
                assert from_node, "No connection found"
                assert isinstance(from_node, list), "Connection must be a list"
                ref_in = from_node[0]["from"]
                lo.remove_node(to_node.name)

        # TO DO, ADD PHY PER CONVERTER
        if not converters:
            converters_first = None
        else:
            converters_first = converters[0]
        in_c, out_c, connect_to_input, xcvr_out = self._draw_phy(
            config, converters_first
        )

        if not system_draw:
            self.ic_diagram_node.add_connection(
                {"from": ref_in, "to": in_c, "rate": clocks["FPGA_REF"]}
            )

        else:
            for converter in converters:
                rcn = f"{self.name}_{converter.name.upper()}_ref_clk"
                assert rcn in clocks, f"Missing clock {rcn}"
                self.ic_diagram_node.add_connection(
                    {"from": ref_in, "to": in_c, "rate": clocks[rcn]}
                )
                if connect_to_input:
                    for c in connect_to_input:
                        self.ic_diagram_node.add_connection(
                            {"from": ref_in, "to": c, "rate": clocks[rcn]}
                        )

        # Delete Transceiver node
        self.ic_diagram_node.remove_child("Transceiver")

        # Connect out_c to JESD204-Link-IP
        if not system_draw:
            self.ic_diagram_node.add_connection(
                {
                    "from": out_c,
                    "to": self.ic_diagram_node.get_child("JESD204-Link-IP"),
                    "rate": clocks["LINK_OUT_REF"],
                }
            )
        else:
            for converter in converters:
                self.ic_diagram_node.add_connection(
                    {
                        "from": out_c,
                        "to": self.ic_diagram_node.get_child("JESD204-Link-IP"),
                        "rate": clocks[
                            f"{self.name}_{converter.name.upper()}_device_clk"
                        ],
                    }
                )

        # Connect device clock to JESD204-Link-IP and JESD204-Transport-IP
        if not system_draw:
            if device_clock_source == "external":
                device_clock = Node("Device Clock", ntype="input")
                lo.add_node(device_clock)
            elif device_clock_source == "link_clock":
                device_clock = out_c
            elif device_clock_source == "ref_clock":
                device_clock = ref_in
            else:
                raise Exception(f"Unknown device clock source: {device_clock_source}")
            self.ic_diagram_node.add_connection(
                {
                    "from": device_clock,
                    "to": self.ic_diagram_node.get_child("JESD204-Link-IP"),
                    "rate": clocks["LINK_OUT_REF"],
                }
            )
            self.ic_diagram_node.add_connection(
                {
                    "from": device_clock,
                    "to": self.ic_diagram_node.get_child("JESD204-Transport-IP"),
                    "rate": clocks["LINK_OUT_REF"],
                }
            )

        else:
            for converter in converters:

                c_name = f"{self.name}_{converter.name.upper()}_device_clk"
                to_node = lo.get_node(c_name)
                # Locate node connected to this one
                from_node = lo.get_connection(to=to_node.name)
                assert from_node, "No connection found"
                assert isinstance(from_node, list), "Connection must be a list"
                lo.remove_node(to_node.name)

                if device_clock_source == "external":
                    device_clock = from_node[0]["from"]
                elif device_clock_source == "link_clock":
                    device_clock = out_c
                elif device_clock_source == "ref_clock":
                    device_clock = ref_in
                else:
                    raise Exception(
                        f"Unknown device clock source: {device_clock_source}"
                    )

                self.ic_diagram_node.add_connection(
                    {
                        "from": device_clock,
                        "to": self.ic_diagram_node.get_child("JESD204-Link-IP"),
                        "rate": clocks[c_name],
                    }
                )
                self.ic_diagram_node.add_connection(
                    {
                        "from": device_clock,
                        "to": self.ic_diagram_node.get_child("JESD204-Transport-IP"),
                        "rate": clocks[c_name],
                    }
                )

        # Connect SYSREF to JESD204-Link-IP
        if not system_draw:
            sysref = Node("SYSREF", ntype="input")
            lo.add_node(sysref)
        else:
            for converter in converters:
                parent = lo.get_node(converter.name.upper())
                to_node = parent.get_child("JESD204 Framer")
                # Locate node connected to this one
                from_node = lo.get_connection(to=to_node.name)
                assert from_node, "No connection found"
                assert isinstance(from_node, list), "Connection must be a list"
                sysref = from_node[0]["from"]
                # lo.remove_node(to_node.name)

        self.ic_diagram_node.add_connection(
            {
                "from": sysref,
                "to": self.ic_diagram_node.get_child("JESD204-Link-IP"),
            }
        )

        # Update with config settings

        # Datapath

        # Get deframer
        if system_draw:
            deframer = lo.get_node("JESD204 Deframer")
            assert deframer, "No JESD204 Deframer found in layout"
            parent = lo.get_node(converter.name.upper())
            framer = parent.get_child("JESD204 Framer")
            assert framer, "No JESD204 Framer found in layout"

            # Replace deframer with a new one inside the FPGA IC diagram
            lo.remove_node("JESD204 Deframer")
            phy_parent = self.ic_diagram_node.get_child("JESD204-PHY-IP")
            new_deframer = Node("DESERIALIZER", ntype="serdes")
            phy_parent.add_child(new_deframer)
            lo.add_connection(
                {
                    "from": xcvr_out,
                    "to": new_deframer,
                    "rate": converter.bit_clock,
                }
            )
            # Add connect for each lane
            for _ in range(converter.L):
                lane_rate = converter.bit_clock
                lo.add_connection(
                    {
                        "from": framer,
                        "to": new_deframer,
                        "rate": lane_rate,
                        "type": "data",
                    }
                )
        if system_draw:
            # Connect DESERIALIZER to link layer decoder
            link_layer = self.ic_diagram_node.get_child("JESD204-Link-IP")
            assert link_layer, "No JESD204-Link-IP found in layout"
            decoder = Node("Link Layer Decoder", ntype="decoder")
            link_layer.add_child(decoder)
            div = 40 if converter.encoding == "8b10b" else 66
            for _ in range(converter.L):
                # Add connect for each lane
                lo.add_connection(
                    {
                        "from": new_deframer,
                        "to": decoder,
                        "rate": converter.bit_clock / div,
                        "type": "data",
                    }
                )

        if system_draw:
            # Add deframer in transport layer
            transport_layer = self.ic_diagram_node.get_child("JESD204-Transport-IP")
            assert transport_layer, "No JESD204-Transport-IP found in layout"
            deframer = Node("JESD204 Deframer", ntype="deframer")
            transport_layer.add_child(deframer)
            # Connect to link layer decoder
            lo.add_connection(
                {
                    "from": decoder,
                    "to": deframer,
                    "rate": converter.bit_clock
                    / (40 if converter.encoding == "8b10b" else 66),
                    "type": "data",
                }
            )
            # Remove connection between link and transport layers
            self.ic_diagram_node.remove_connection(
                from_s=link_layer.name, to=transport_layer.name
            )

        return lo.draw()



"""AD9144 high speed DAC clocking model."""

from typing import Any, Dict, List, Union

import numpy as np

from adijif.converters.ad9144_bf import ad9144_bf

from ..solvers import CpoSolveResult  # noqa: I202


def _convert_to_config(
    L: Union[int, float],
    M: Union[int, float],
    F: Union[int, float],
    S: Union[int, float],
    # HD: Union[int, float],
    N: Union[int, float],
    Np: Union[int, float],
    # CS: Union[int, float],
    DualLink: bool,
) -> Dict:
    # return {"L": L, "M": M, "F": F, "S": S, "HD": HD, "N": N, "Np": Np, "CS": CS}
    return {
        "L": L,
        "M": M,
        "F": F,
        "S": S,
        "HD": 1 if F == 1 else 0,
        "Np": Np,
        "CS": 0,
        "DualLink": DualLink,
    }


# IF F==1, HD=1, otherwise 0
quick_configuration_modes = {
    str(0): _convert_to_config(DualLink=False, M=4, L=8, S=1, F=1, N=16, Np=16),
    str(1): _convert_to_config(DualLink=False, M=4, L=8, S=2, F=2, N=16, Np=16),
    str(2): _convert_to_config(DualLink=False, M=4, L=4, S=1, F=2, N=16, Np=16),
    str(3): _convert_to_config(DualLink=False, M=4, L=2, S=1, F=4, N=16, Np=16),
    str(4): _convert_to_config(DualLink=False, M=2, L=4, S=1, F=1, N=16, Np=16),
    str(5): _convert_to_config(DualLink=False, M=2, L=4, S=2, F=2, N=16, Np=16),
    str(6): _convert_to_config(DualLink=False, M=2, L=2, S=1, F=2, N=16, Np=16),
    str(7): _convert_to_config(DualLink=False, M=2, L=1, S=1, F=4, N=16, Np=16),
    # 8 is missing in datasheet
    str(9): _convert_to_config(DualLink=False, M=1, L=2, S=1, F=1, N=16, Np=16),
    str(10): _convert_to_config(DualLink=False, M=1, L=1, S=1, F=2, N=16, Np=16),
}

quick_configuration_modes = {
    **quick_configuration_modes,
    **{
        str(4)
        + "-DL": _convert_to_config(DualLink=False, M=2, L=4, S=1, F=1, N=16, Np=16),
        str(5)
        + "-DL": _convert_to_config(DualLink=False, M=2, L=4, S=2, F=2, N=16, Np=16),
        str(6)
        + "-DL": _convert_to_config(DualLink=False, M=2, L=2, S=1, F=2, N=16, Np=16),
        str(7)
        + "-DL": _convert_to_config(DualLink=False, M=2, L=1, S=1, F=4, N=16, Np=16),
        # 8 is missing in datasheet
        str(9)
        + "-DL": _convert_to_config(DualLink=False, M=1, L=2, S=1, F=1, N=16, Np=16),
        str(10)
        + "-DL": _convert_to_config(DualLink=False, M=1, L=1, S=1, F=2, N=16, Np=16),
    },
}


class ad9144(ad9144_bf):
    """AD9144 high speed DAC model.

    This model supports both direct clock configurations and on-board
    generation

    Clocking: AD9144 has directly clocked DAC that have optional input dividers.
    The sample rate can be determined as follows:

        baseband_sample_rate = (input_clock / input_clock_divider) / datapath_decimation
    """

    name = "AD9144"
    converter_type = "DAC"

    # JESD parameters
    _jesd_params_to_skip_check = ["DualLink", "K"]
    DualLink = False
    available_jesd_modes = ["jesd204b"]
    K_available = [16, 32]
    _K = 32  # Valid for all cases pg 31 of datasheet
    L_available = [1, 2, 4, 8]
    M_available = [1, 2, 4, 4]
    N_available = [*range(7, 16 + 1)]
    Np_available = [8, 16]
    F_available = [1, 2, 4, 8, 16]
    CS_available = [0, 1, 2, 3]
    CF_available = [0]
    S_available = [1, 2]

    # Clock constraints
    clocking_option_available = ["integrated_pll", "direct"]
    _clocking_option = "direct"

    converter_clock_min = 420e6  # checked by dac
    converter_clock_max = 2.8e9  # checked by dac

    bit_clock_min_available = {"jesd204b": 1.44e9}  # checked by jesd
    bit_clock_max_available = {"jesd204b": 12.4e9}  # checked by jesd

    sample_clock_min = 1.44e9 / 40  # checked by jesd
    sample_clock_max = 1.06e9  # checked by jesd

    interpolation_available = [1, 2, 4, 8]
    _interpolation = 1

    quick_configuration_modes = {"jesd204b": quick_configuration_modes}

    # Input clock requirements
    input_clock_divider_available = [
        1,
        2,
        4,
        8,
        16,
    ]  # only used with integrated PLL for now

    # Integrated PLL limits
    pfd_min = 35e6
    pfd_max = 80e6

    # With integrated PLL
    input_clock_min = 35e6
    input_clock_max = 1e9

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize AD9144 class.

        Objects will default to mode 0x88 with 1e9 sample_clock.

        Args:
            *args (Any): Pass through arguments
            **kwargs (Any): Pass through keyword arguments
        """
        # Set default mode
        self.set_quick_configuration_mode(str(9))
        self.sample_clock = 1e9
        super().__init__(*args, **kwargs)
        # self._init_diagram()

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal converter configuration and output clock definitions
        leading to connected devices (clock chips, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration
        """
        config: Dict = {
            "clocking_option": self.clocking_option,
            "interpolation": self._interpolation,
        }
        if self.clocking_option == "direct":
            return config

        if self.solver == "CPLEX":
            if solution:
                self.solution = solution
            config.update(
                {
                    "BCount": self._get_val(self.config["BCount"]),
                    "ref_div_factor": self._get_val(self.config["ref_div_factor"]),
                    "lo_div_mode": np.log2(
                        self._get_val(self.config["lo_div_mode_p2"])
                    ),
                }
            )
        return config

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Returns:
            List[str]: List of strings of clock names in order
        """
        # clk = (
        # "AD9144_dac_clock" if self.clocking_option == "direct" else "AD9144_pll_ref"
        # )
        clk = "ad9144_ref_clk"
        return [clk, "ad9144_sysref"]

    def _check_valid_internal_configuration(self) -> None:
        mode = self._check_valid_jesd_mode()
        # cfg = self.quick_configuration_modes[self.jesd_class][mode]

        if mode in ["0", "4", "9"]:
            assert self.K == 32, "K must be 32 for JESD mode 0, 4, or 9"

    def _pll_config(self) -> Dict:
        dac_clk = self.interpolation * self.sample_clock
        self.config["dac_clk"] = self._convert_input(dac_clk, "dac_clk")

        self.config["BCount"] = self._convert_input([*range(6, 127 + 1)], name="BCount")

        # Datasheet says refclk div can support 32 but tables do not reflect this and
        # a div of 32 would put you under supported range

        if self.solver == "gekko":
            self.config["ref_div_factor"] = self.model.sos1(
                self.input_clock_divider_available
            )
            # self.config["ref_div_factor_i"] = self.model.Var(
            #     integer=True, lb=0, ub=4, value=4
            # )
            # self.config["ref_div_factor"] = self.model.Intermediate(
            #     2 ** (self.config["ref_div_factor_i"])
            # )

            self.config["ref_clk"] = self.model.Var(
                integer=False, lb=35e6, ub=1e9, value=35e6
            )
        elif self.solver == "CPLEX":
            self.config["ref_div_factor"] = self._convert_input(
                self.input_clock_divider_available, "ref_div_factor"
            )

            self.config["ref_clk"] = (
                self.config["dac_clk"]
                * self.config["ref_div_factor"]
                / self.config["BCount"]
                / 2
            )

        if dac_clk > 2800e6:
            raise Exception("DAC Clock too fast")
        elif dac_clk >= 1500e6:
            self.config["lo_div_mode_p2"] = self._convert_input(
                2 ** (1 + 1), name="lo_div_mode_p2"
            )
        elif dac_clk >= 720e6:
            self.config["lo_div_mode_p2"] = self._convert_input(
                2 ** (2 + 1), name="lo_div_mode_p2"
            )
        elif dac_clk >= 420e6:
            self.config["lo_div_mode_p2"] = self._convert_input(
                2 ** (3 + 1), name="lo_div_mode_p2"
            )
        else:
            raise Exception("DAC Clock too slow")

        self.config["vco"] = self._add_intermediate(
            self.config["dac_clk"] * self.config["lo_div_mode_p2"]
        )

        self._add_equation(
            [
                self.config["ref_div_factor"] * self.pfd_min < self.config["ref_clk"],
                self.config["ref_div_factor"] * self.pfd_max > self.config["ref_clk"],
                self.config["ref_clk"] >= self.input_clock_min,
                self.config["ref_clk"] <= self.input_clock_max,
            ]
        )

        if self.solver == "gekko":
            self._add_equation(
                [
                    self.config["ref_clk"] * 2 * self.config["BCount"]
                    == self.config["dac_clk"] * self.config["ref_div_factor"]
                ]
            )

        return self.config["ref_clk"]

    def get_required_clocks(self) -> List:
        """Generate list required clocks.

        For AD9144 this will contain [converter clock, sysref requirement SOS]

        Returns:
            List: List of dictionaries of solver components
        """
        self.config = {}

        self.config["lmfc_divisor_sysref"] = self._convert_input(
            [*range(1, 20 + 1)], name="lmfc_divisor_sysref"
        )

        self.config["sysref"] = self._add_intermediate(
            self.multiframe_clock / (self.config["lmfc_divisor_sysref"])
        )

        if self.clocking_option == "direct":
            clk = self.sample_clock * self.interpolation
            # LaneRate = (20 × DataRate × M)/L
            assert (
                self.bit_clock == (20 * self.sample_clock * self.M) / self.L
            ), "AD9144 direct clock requirement invalid"
        else:
            # vco = dac_clk * 2^(LO_DIV_MODE + 1)
            # 6 GHz <= vco <= 12 GHz
            # BCount = floor( dac_clk/(2 * ref_clk/ref_div ) )
            # 5 <= BCount <= 127
            # ref_div = 2^ref_div_mode = 1,2,4,8,16
            clk = self._pll_config()  # type: ignore

        # Objectives
        # self.model.Obj(self.config["sysref"])  # This breaks many searches
        # self.model.Obj(-1*self.config["lmfc_divisor_sysref"])

        return [clk, self.config["sysref"]]



# flake8: noqa
import os
import pathlib
import runpy

import pytest

ignore = ["daq2_rx_dt_example.py", "hmc7044_example.ipynb"]

examples_dir = os.path.join(pathlib.Path(__file__).parent.parent, "examples")
examples_dir = pathlib.Path(examples_dir)
examples = [str(p) for p in examples_dir.glob("*.py")]
filenames = [os.path.basename(i) for i in examples]
filenames = [i for i in filenames if i not in ignore]
examples = [os.path.join(examples_dir, i) for i in filenames]


@pytest.mark.parametrize("script", examples)
def test_example_execution(script):
    runpy.run_path(script)



# flake8: noqa
import pprint

import pytest

import adijif


def test_adf4382_integer_datasheet_example():

    # Reference example from AD9084 system
    ######################
    pll = adijif.adf4382()

    ref_in = int(125e6 / 2)
    output_clocks = int(8e9)
    pll.n = 128
    pll.d = 1
    pll.o = 2

    # adf4382 spi3.0: VCO=16000000000 PFD=62500000 RFout_div=2 N=128 FRAC1=0 FRAC2=0 MOD2=0
    o = 2  #
    n = 128  #
    d = 1  # 1
    r = 1
    vco_ref = ref_in * d * n * o / r
    pfd_ref = ref_in * d / r

    pll.set_requested_clocks(ref_in, output_clocks)

    pll.solve()

    cfg = pll.get_config()

    pprint.pprint(cfg)

    pfd = ref_in * cfg["d"] / cfg["r"]
    vco = pfd * cfg["n"] * cfg["o"]
    rf_out = vco / cfg["o"]

    print(f"PFD: {pfd}")
    print(f"VCO: {vco}")
    print(f"RF Out: {rf_out}")

    assert rf_out == output_clocks, "Output frequency does not match requested"
    assert cfg["o"] == o, "Output divider does not match datasheet example"
    assert cfg["n"] == n, "N divider does not match datasheet example"
    assert cfg["d"] == d, "D divider does not match datasheet example"
    assert cfg["r"] == r, "R divider does not match datasheet example"
    assert vco == vco_ref, "VCO frequency does not match datasheet example"
    assert pfd == pfd_ref, "PFD frequency does not match datasheet example"


@pytest.mark.parametrize(
    "ref_in", [int(125e6), int(125e6 / 2), int(125e6 / 4), int(125e6 / 8)]
)
def test_adf4382_auto_20g(ref_in):
    pll = adijif.adf4382()
    pll.mode = "integer"

    output_clocks = int(20e9)

    pll.set_requested_clocks(ref_in, output_clocks)

    pll.solve()

    cfg = pll.get_config()
    pprint.pprint(cfg)

    assert (
        output_clocks == cfg["rf_out_frequency"]
    ), "Output frequency does not match requested"


def test_adf4382_frac_datasheet_org():
    ref_in = int(250e6)

    pll = adijif.adf4382()
    pll.mode = "fractional"
    pll.require_phase_sync = False
    pll.d = 1
    pll.o = 1
    pll.r = 1
    pll.n = 80

    # val1 = (17716740 + 1500000/15625000)/ 33554432

    # Since the values are a bit arbitrary, we can just set a few of them to the
    # values in the datasheet
    pll._frac1_min_max = [17716740, 17716740]
    pll._frac2_min_max = [1500000, 1500000]

    output_clocks = int(20.132e9)

    pll.set_requested_clocks(ref_in, output_clocks)

    pll.solve()

    cfg = pll.get_config()
    pprint.pprint(cfg)

    n = cfg["n_int"] + (cfg["n_frac1w"] + cfg["n_frac2w"] / cfg["MOD2"]) / pll._MOD1
    rf_out = ref_in * cfg["d"] * n / cfg["r"]
    print(f"RF Out: {rf_out}")

    # Verify with high precision
    import mpmath as mp

    mp.mp.dps = 50
    frac1 = mp.mpf(cfg["n_frac1w"])
    frac2 = mp.mpf(cfg["n_frac2w"])
    MOD2 = mp.mpf(cfg["MOD2"])
    MOD1 = mp.mpf(33554432)
    n_int = mp.mpf(cfg["n_int"])
    rf_out_hp = ref_in * cfg["d"] * (n_int + (frac1 + frac2 / MOD2) / MOD1) / cfg["r"]
    print(f"RF Out HP: {rf_out_hp}")
    assert rf_out_hp == pytest.approx(rf_out, rel=1e-6)

    assert (
        output_clocks == cfg["rf_out_frequency"]
    ), "Output frequency does not match requested"
    assert cfg["n_frac1w"] == 17716740, "Fractional part 1 does not match datasheet"
    assert cfg["n_frac2w"] == 1500000, "Fractional part 2 does not match datasheet"
    assert cfg["MOD2"] == 15625000, "MOD2 does not match datasheet"


def test_adf4382_frac_datasheet_auto():
    ref_in = int(250e6)

    pll = adijif.adf4382()
    pll.mode = ["fractional", "integer"]

    output_clocks = int(20.132e9)
    pll.set_requested_clocks(ref_in, output_clocks)

    pll.solve()

    cfg = pll.get_config()
    pprint.pprint(cfg)

    n = cfg["n_int"] + (cfg["n_frac1w"] + cfg["n_frac2w"] / cfg["MOD2"]) / pll._MOD1
    rf_out = ref_in * cfg["d"] * n / cfg["r"]
    print(f"RF Out: {rf_out}")

    # Verify with high precision
    import mpmath as mp

    mp.mp.dps = 50
    frac1 = mp.mpf(cfg["n_frac1w"])
    frac2 = mp.mpf(cfg["n_frac2w"])
    MOD2 = mp.mpf(cfg["MOD2"])
    MOD1 = mp.mpf(33554432)
    n_int = mp.mpf(cfg["n_int"])
    rf_out_hp = ref_in * cfg["d"] * (n_int + (frac1 + frac2 / MOD2) / MOD1) / cfg["r"]
    print(f"RF Out HP: {rf_out_hp}")
    assert rf_out_hp == pytest.approx(rf_out, rel=1e-6)

    assert (
        output_clocks == cfg["rf_out_frequency"]
    ), "Output frequency does not match requested"
